// CmdString.cpp: implementation of the CCmdString class.
//
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "CmdString.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//=========================================================================//
//                                                                         //
//            Encodes a sequence of deformation the Rubik's Cube           //
//                                                                         //
t_Mask Command(short axe,     // 0..2 => x, y or z                         //
               short plan,    // 0..2 => No slice, 3 => full Cube          //
               bool neg)      // CCW rotation                              //
//=========================================================================//
{
   t_Mask m = ((plan & 3)==3)? (0x0001 << (axe * 2))|0x0100:
                   ((0x603 & (plan + 0x0601)) << (axe << 1));
   return neg? m | 0x8000: m;
}

//=========================================================================//
//                                                                         //
//   Extract the rotation axis of a deformation order of the Rubik's Cube  //
//                                                                         //
short GetAxe(t_Mask cmd)                                                   //
//=========================================================================//
{
   short axe;
   for (axe = 0; axe < 3; axe++, cmd >>= 2) if (cmd & 3) break;
   return axe;
}

//=========================================================================//
//                                                                         //
//                      Calculate the command mask                         //
//                                                                         //
t_Mask CmdCubeSet(t_Command cmd)                                           //
//=========================================================================//
{
   BYTE face = (cmd & 0x1F) % 6, or = (cmd & 0x1F) / 6, axe = (face / 2) & 3;
   t_Mask out;

   if ((cmd & 0x80) || cmd < 6) return 0;             // multiple rotations
   else if (cmd & 0x40) out = (1 << (axe*2))|0x0100;  // full cube
   else if (cmd & 0x20) out = 0x0602 << (axe*2);      // slice 2
   else if (face & 1)   out = 0x0601 << (axe*2);      // slice 1
   else                 out = 0x0603 << (axe*2);      // slice 3

   if (face & 1)        out ^= 0x8000;
   if (or == 3)         out ^= 0x8000;

   return out;
}

//=========================================================================//
//                                                                         //
//                   Explode complex orders of rotation                    //
//                                                                         //
t_Command CmdTranslate(t_Command &cmd)                                     //
//=========================================================================//
{
   t_Command c = cmd;
   if (c & 0x80) {
      if (c & 0x40) {
         // CCW order
         for (int i = 4; i >= 0; i -= 2) {
            if (c = ((cmd >> i)& 3)) {
               cmd &= ~(3 << i);
               return i + c * 6 + 0x41;
            }
         }
      } else {
         // CW order
         for (int i = 0; i < 6; i += 2) {
            if (c = ((cmd >> i)& 3)) {
               cmd &= ~(3 << i);
               return i + c * 6 + 0x40;
            }
         }
      }
   }
   cmd = 0;
   return c;
}

//=========================================================================//
static bool SearchCommand(const CString &msg,   // line in a string
                          int ptr)     // begining of the line in a string
{
   for(int i = ptr; i < msg.GetLength(); i++) {
      static const CString code("\r\n/rlbfudRLBFUD"); // char to detect
      TCHAR c = msg.GetAt(i);
      int ix = code.Find(c);
      if (ix < 0) continue;
      else if (ix < 3) break;          // end of line or comments
      else return true;                // ok, char of commands
   }
   return false;
}
//=========================================================================//
//                                                                         //
//                   Invalidate commands line in a text                    //
//                                                                         //
CString InvalidCommands(const CString &string)                             //
//=========================================================================//
{
   CString out = string;

   for (int i = string.GetLength() - 1; --i >= 0; ) {
      if (string.Mid(i, 2) == "\r\n") {
         // new line detected
         if (SearchCommand(string, i + 2))
            out.Insert(i + 2, "/ ");   // insert a comments marker
      }
   }

   // search at the begin of the string
   if (SearchCommand(string, 0))
      out.Insert(0, "/ ");

   // terminate the text lines
   out.TrimLeft(), out.TrimRight();
   if (!out.IsEmpty() && out.Right(2) != "\r\n") out += "\r\n";

   return out;
}

//=========================================================================//
//static const CString acode("rlbfudRLBFUD>*/\n");
static const CString acode("rlbfudRLBFUD>*/\n");   // command chars
//=========================================================================//
//                                                                         //
//                Convert ascii commands in order codes                    //
//                                                                         //
CCmdString CCmdString::CmdExtract(const CString &src, // source input      //
                                  int cur,   // first item to be processed //
                                  int *pLast)   // end index out           //
//=========================================================================//
{
   int len = src.GetLength(), beg = cur, flag = 0;
   bool frst = true;

//   Empty();
   // the command interpreter, idles up to encounter an order
   while (cur < len) {
      TCHAR cmd = src.GetAt(cur++);
      int code = acode.Find(cmd);    // code( -1..15 )

      switch (code) {
      case 12 : if (flag == 0) { flag = 0x20; continue; } // '>'
      case 13 : if (flag == 0) { flag = 0x40; continue; } // '*'
      case 16 :
      case -1 : break;                    // invalid char, next !

      case 14 :   // '/' start of comments, search end of line
         while (cur < len && src.GetAt(cur++) != TCHAR('\n')) ;
      case 15 :   //'\n' new line
         if (pLast) goto end; else if (!frst) *this += TCHAR(1), frst = true;
         break;
        
      default :
         ASSERT(code >= 0 && code < 12);
         // is a command, the code is between 0 and 11

         if (code < 6) code += 18;        // CCW (lowercase)

         // post fixed orders
         if (cur < len) {
            TCHAR u = src.GetAt(cur++);   // next char

            if (u == TCHAR('\'') || u == TCHAR('') || u == TCHAR('3'))
               code = (code + 12) % 24;   // CCW 
            else if (u == TCHAR('2') || cmd == u)
               code = (code % 6) + 12;    // 180 degrees
            else
               cur--;
         }

         // generate the new command
         *this += TCHAR(code | flag);
         frst = false;
      }
      flag = 0;                           // dsarmer le prfixe
      beg = cur;                          // new command begin
   }
end:;
//   if (Right(1) != 1) *this += TCHAR(1);
   if (pLast) *pLast = cur;
   return *this;
}

//=========================================================================//
//                                                                         //
//              Convert an ascii command in the order code                 //
//                                                                         //
t_Command CCmdString::CmdExtractOne(const CString &src,// source input     //
                                 int cur,    // first item to be processed //
                                 int &last)     // end index out           //
//=========================================================================//
{
   int len = src.GetLength(), flag = 0;

   while (cur < len) {
      CString &cmd = src.Mid(cur++, 1);
      int code = acode.FindOneOf(cmd);    // code( -1..15 )

      switch (code) {
      case 12 : if (flag == 0) { flag = 0x20; break; }   // '>'
      case 13 : if (flag == 0) { flag = 0x40; break; }   // '*'
      case -1 : last = cur; return 0;     // invalid char, we pass !

      case 14 : while (cur < len && src.GetAt(cur++) != TCHAR('\n')) ;
      case 15 : last = cur; return 0;     // comments, we pass !

      default :
         ASSERT(code >= 0 && code < 12);
         // is a command, the code is between 0 and 11

         if (code < 6) code += 18;        // CCW (lowercase)

         // post fixed orders
         if (cur < len) {
            TCHAR u = src.GetAt(cur++);   // next char

            if (u == TCHAR('\'') || u == TCHAR('') || u == TCHAR('3'))
               code = (code + 12) % 24;   // CCW 
            else if (u == TCHAR('2') || cmd == u)
               code = (code % 6) + 12;    // 180 degrees
            else
               cur--;
         }

         last = cur;
         // generate the new command
         return code | flag;
      }
   }
   last = cur;
   return 0;
}

//=========================================================================//
//                                                                         //
//                 Convert command in the ascii codes                      //
//                                                                         //
CString CCmdString::CmdConvert(bool cr, const CString &info) const         //
//=========================================================================//
{
   static const int lengthMax = 75;
   int len = GetLength(), n, cnt = 0, cntq = 0;
   bool lcr = cr;
   CString text;

   
   for (int i = 0, ol = 0; i < len; i++) {
      t_Command c = GetAt(i);
      CString cmd;

      if (c < 6) {
         // special codes

         switch (c) {
         case 1 :
            if (text.GetLength() != ol) {
               if (lcr && cnt > 0) {
                  CString tmp;

                  // add commands count at end of string
                  tmp.Format(" (%dq/%d)%s", cntq, cnt, info);
                  text += tmp;
                  lcr = false;
               }
               text += CmdConvertOneCmd(c), ol = text.GetLength();
            }
            break;
         case 2 : cmd = Mid(i + 1, 8), i += 8; break;
         case 5 : cmd = Mid(i + 1, 27), i += 27; break;
         }


      } else if (c & 0x80) {
         // Treatment of combined rotations along the three axes

         cmd = TCHAR('[');
         do {
            BYTE cx = CmdTranslate(c);

            if (cx)
               cmd += CString(TCHAR('*')) + CmdConvertOneCmd(cx);
         } while (c != 0);
         cmd += TCHAR(']');
         cnt++;

      } else {
         // Treatment of slide rotations
         static const TCHAR pre[] = ">* ";


         ASSERT((c & 0x9F) < 24);
         if (n = (c >> 5)) cmd = pre[n - 1];
         cmd += CmdConvertOneCmd(c);
         if (c / 6 == 2) cntq += 2; else cntq++;
         cnt++;
      }

      if (text.GetLength() + cmd.GetLength() > ol + lengthMax)
         text += CmdConvertOneCmd(1), ol = text.GetLength();
      else if (text.GetLength() != ol)
         text += TCHAR(' ');

      text += cmd;
   }

   if (cr && text.IsEmpty()) {
      text = "/ Empty";
   } else if (lcr && cnt > 0) {
      CString tmp;

      // add commands count at end of string
      tmp.Format(" (%dq/%df)", cntq, cnt);
      text += tmp;
      if (!info.IsEmpty()) text += CString(' ') + info;
   }

   return text;
}

//=========================================================================//
//                                                                         //
t_Command CCmdString::toCommand(BYTE axe, BYTE cubeNo)                     //
//=========================================================================//
{
//        no Cubes  24 25 26          axe x, R - L
//          couche  21  U 23          axe y, B - F
//      superieure  18 19 20          axe z, U - D
//
//          couche  15  B 17
//         mediane   L 13  R
//                   9  F 11
//
//          couche   6  7  8
//      inferieure   3  D  5
//                   0  1  2
//   typedef enum {R,L,B,F,U,D};            // faces names
//   using namespace CCmdString;

   static BYTE scmdTbl[3][27] = {
//  axe x, R - L
//  0  1      2   3  4      5   6  7      8   9  10     11  12 13
   {cL,cL|cSc,cRr,cL,cL|cSc,cRr,cL,cL|cSc,cRr,cL,cL|cSc,cRr,0, 0,
//  14 15 16     17  18 19     20  21 22     23  24 25     26
    0, cL,cL|cSc,cRr,cL,cL|cSc,cRr,cL,cL|cSc,cRr,cL,cL|cSc,cRr},
//  axe y B - F
//  0  1  2  3      4      5      6   7   8   9  10 11 12     13
   {cF,cF,cF,cF|cSc,cF|cSc,cF|cSc,cBr,cBr,cBr,cF,0, cF,cF|cSc,0,
//  14     15  16 17  18 19 20 21     22     23     24  25  26
    cF|cSc,cBr,0, cBr,cF,cF,cF,cF|cSc,cF|cSc,cF|cSc,cBr,cBr,cBr},
//  axe z U - D
//  0  1  2  3  4  5  6  7  8  9      10     11     12     13
   {cD,cD,cD,cD,0, cD,cD,cD,cD,cD|cSc,cD|cSc,cD|cSc,cD|cSc,0,
//  14     15     16     17     18  19  20  21  22 23  24  25  26
    cD|cSc,cD|cSc,cD|cSc,cD|cSc,cUr,cUr,cUr,cUr,0, cUr,cUr,cUr,cUr}
   };

   return (cubeNo < 27)? scmdTbl[axe][cubeNo]: 0;
}

//=========================================================================//
//                                                                         //
//                Convert of an command line to debug in hexa              //
//                                                                         //
CString CCmdString::CmdHexaDump() const                                    //
//=========================================================================//
{
   CString text, tmp;
   t_Command cmd;

   for (int i = 0; i < GetLength(); i++) {
      if ((cmd = GetAt(i)) == 1) text += "\r\n";
      else tmp.Format(" %02x", cmd), text += tmp;
   }

   return text;
}

CString AddStringWithCr(const CString &msg1, const CString &msg2,
                        const CString cr)
{
   int lCr = cr.GetLength();
   if (msg1.IsEmpty()) return msg2;
   if (msg2.IsEmpty()) return msg1;

   return (msg1.Right(lCr) == cr || msg2.Left(lCr) == cr)? msg1 + msg2:
                                  msg1 + cr + msg2;
}

//=========================================================================//
inline static TCHAR Combine(TCHAR a, TCHAR b)
   { return (((a &= 0x1F) / 6 + (b & 0x1F) / 6) & 3) * 6 + a % 6; }
inline static TCHAR getFace(TCHAR a)
   { return (a & 0x1F) % 6 + (a & 0x60); }
inline static TCHAR getAxe(TCHAR a)
   { return ((a & 0x1F) % 6) / 2; }
//=========================================================================//
//                                                                         //
//                        Optimize a command line                          //
//                                                                         //
CCmdString CCmdString::CmdOptimize()                                       //
//=========================================================================//
{
   static const t_Command t_CNull = t_Command(-1);
   t_Command f2 = t_CNull, f1 = t_CNull, f0, o, cur;
   int len = GetLength(), i, j;

   for (i = j = 0; i < len; i++) {
      cur = GetAt(i);     // first order

      if (cur < 6 || (cur & 0x80)) {
         ASSERT(cur != 5);
         f2 = f1 = t_CNull, SetAt(j++, cur); continue;
      }

      f0 = getFace(cur);       // face one
      if (f0 == f1) {
         // two consecutive orders are for the same face (and even Axis)
         o = Combine(cur, GetAt(j - 1));
         if (o < 6) {
            // two consecutive orders vanish
            --j, f1 = f2, f2 = (j > 1)? getFace(GetAt(j - 2)): t_CNull;
         } else {
            // two consecutive orders are combined into one
            SetAt(j - 1, o | (cur & 0x60));
         }
      } else if (((f0 ^ f1)& 0x7E) != 0) {
         // this order is not in the same axis as the previous
         f2 = t_CNull, f1 = f0, SetAt(j++, cur);
      } else if (f0 == f2) {
         // three consecutive orders are in the same axis and the first and
         // third relate to the same face
         o = Combine(cur, GetAt(j - 2));
         if (o < 6) {
            // the first and third vanish
            SetAt(j - 2, GetAt(j - 1));
            --j, f1 = f2, f2 = (j > 0)? getFace(GetAt(j - 1)): t_CNull;
         } else {
            // the first and third are combined in the only one
            SetAt(j - 2, o | (cur & 0x60));
         }
      } else {
         // This order is in the same axis as the previous one but not the
         // same face.
         f2 = f1, f1 = f0, SetAt(j++, cur);
         // Note : insofar as there are only two sides per axis, as three
         // levels to follow along the same axis, they must lead
         // optimization, because both are the same face.
      }
      
      // R,  L,  B,  F,  U,  D
      if ( f2 != t_CNull && getAxe(f1) == getAxe(f2)
        && (f1 & 0x7) > (f2 & 0x7) ) {
         // permutes the facets movements as they relate to opposite sides
         cur = GetAt(j - 1), SetAt(j - 1, GetAt(j - 2)), SetAt(j - 2, cur);
         f0 = f1, f1 = f2, f2 = f0;
      }
   }

   if (j < i) Delete(j, i - j);

   return *this;
}

/////////////////////////////////////////////////////////////////////////////
// Test sequences of the optimizer
#ifdef _DEBUG
static TCHAR *testSeq[] = {
   // R,  L,  B,  F,  U,  D
   // no modification
   "L F B R D",
   "L F B R D",
   // basic functions
   "L F F B",        // - Convert two identical quarter turn commands in the
   "L F2 B",         // half turn order
   // permutes sides
   "L B' F D",       // - to head faces U, R, F when successive faces are
   "L F B' D",       // opposite (to facilate compare sequences).
   //
   "L F B' D",       // - idem, no modification
   "L F B' D",
   // all cases
   "R B B L", 
   "R B2 L", 
   //
   "U D' R L'",
   "D' U L' R",
   //
   "U' D R R L2",
   "D U' L2 R2",
   //
   NULL
};

void optimizerTest()
{
   for (TCHAR **pTest = testSeq; *pTest; pTest += 2) {
      CCmdString cmd;
      CString msg;

      cmd.CmdExtract(*pTest); // read command string
      cmd.CmdOptimize();
      msg = cmd.CmdConvert();

      TRACE(*pTest);
      if (msg.Compare(*(pTest+1))== 0) TRACE(" : ok\n");
      else  TRACE("\n%s : error\n", msg);
   }
}
#else
void optimizerTest() {}
#endif
/////////////////////////////////////////////////////////////////////////////

