// CmdString.h: interface for the CCmdString class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_CMDSTRIN_H__905B8C71_EBA7_434F_B962_E9A453884CDF__INCLUDED_)
#define AFX_CMDSTRIN_H__905B8C71_EBA7_434F_B962_E9A453884CDF__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


/////////////////////////////////////////////////////////////////////////////
// The position of the centers of elementary cubes of a Rubik's Cube can be
// represented by a vector of three values representing the components x, y
// and z. Each of these components taking only three values -1, 0 and 1.
typedef unsigned short t_Mask;
   // cube position :
   //    position Cube(bits9..14) | orientation Cube(bits0..5)
   //      posCube 2 bits per axe
   //      orCube : no 1 to 24 include
   // or commande : 
   //    mask(bits9..15) | rep(bit8) | plan(bits0..5)

// calculate the position of a cube 0..26
inline t_Mask cubePos(BYTE n)
   { return ((n%3 + 1)<< 9)|(((n%9)/3 + 1) << 11)|(((n%27)/9 + 1)<< 13); } 
// Convert the position of a cube in his number 0 to 26
inline BYTE cubeNo(t_Mask c)
   { return BYTE(((c >> 9)&3) + ((c >> 11)&3)*3 + ((c >> 13)&3)*9 - 13); }

// Convert a char vector[3] in a t_Mask position
inline t_Mask vectPack(const char v[3])
   { return ((v[0] + 2)<< 9)|((v[1] + 2)<< 11)|((v[2] + 2)<< 13); }

// Convert a char vector[3] in a number 0 to 26
inline BYTE vPack(const char v[3])
   { return v[0] + v[1] * 3 + v[2] * 9 + 13; }
// Convert a number 0 to 26 in a char vector[3]
inline void vUnPack(BYTE n, char v[3])
   { v[0] = n % 3 - 1, v[1] = (n / 3)% 3 - 1, v[2] = (n / 9)% 3 - 1; }

// test cube positions belonging to a plane 
inline bool CubeChk(t_Mask cube, t_Mask m)
   { WORD c = (m & 0x8000)? 0x2A00: 0x7E00; return ((cube ^ c)& m & 0x7E00)== 0; }

inline CString binConvert(BYTE d)
   { CString msg; while (d) msg += CString(TCHAR('0'+(d & 1))), d >>= 1; return msg; }  

extern t_Mask Command(short axe, short plan, bool neg);
extern short GetAxe(t_Mask cmd);


/////////////////////////////////////////////////////////////////////////////
typedef BYTE t_Command;

extern t_Mask CmdCubeSet(t_Command cmd);
inline int CmdCubeCount(t_Command cmd) { return ((cmd & 0x1F) / 6 == 2)? 2: 1; }
extern t_Command CmdTranslate(t_Command &cmd);

extern CString InvalidCommands(const CString &string);

extern CString AddStringWithCr(const CString &msg1, const CString &msg2,
                               const CString cr = CString("\r\n"));

/////////////////////////////////////////////////////////////////////////////
// Reading and converting a command line
class CCmdString : public CString
{
public:
   CCmdString() : CString() {};
   CCmdString(t_Command cmd) : CString(TCHAR(cmd)) {}
   CCmdString(const t_Command table[], int length)
      : CString((const TCHAR*)table, length) {}

   typedef enum {
      R,  L,  B,  F,  U,  D,     // faces names (x=> R-L, y => B-F, z=> U-D
      cR, cL, cB, cF, cU, cD,    // faces rotation CW 
      cR2,cL2,cB2,cF2,cU2,cD2,   // faces rotation 180 degrees
      cRr,cLr,cBr,cFr,cUr,cDr,   // faces rotation CCW
      cSc=0x20,                  // turn the slice of middle
      cFull=0x40,                // rotate the whole cube
      cRot=0x80,CR=1,Init=5
   } Move;

   inline CString CmdConvertOneCmd(t_Command c) const
      {  static const TCHAR* acode[24] = {
            "",  "\r\n",":",  "",   "#",  "@",
            "R",  "L",  "B",  "F",  "U",  "D",
            "R2", "L2", "B2", "F2", "U2", "D2",
            "R'", "L'", "B'", "F'", "U'", "D'" };
         return acode[c & 0x1F]; }

   CString CmdHexaDump() const;

   CCmdString CmdExtract(const CString &src, int first = 0, int *pLast = NULL);
   static t_Command CmdExtractOne(const CString &src, int first, int &last);

   CCmdString CmdOptimize();
   CString CmdConvert(bool text = false, const CString &info = "") const;

   static t_Command toCommand(BYTE axe, BYTE cubie);
};

extern void optimizerTest();

#endif // !defined(AFX_CMDSTRIN_H__905B8C71_EBA7_434F_B962_E9A453884CDF__INCLUDED_)
