// Colors.cpp: implementation of the Colors utilities.
//
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Colors.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/////////////////////////////////////////////////////////////////////////////
//
// Facets colors of the Rubik's Cube

static const COLORREF defaultColors[8] = {
// standards colors     color    side   face        couleur   code
// ----------------  ----------  -----  ----------  --------  ----
   RGB(255,128,0),   // ORANGE   Right  Droite      O orange  0
   RGB(255,0,0),     // RED      Left   Gauche      R rouge   1
   RGB(0,192,0),     // GREEN    Back   Posterieur  G vert    2
   RGB(0,0,255),     // BLUE     Front  Avant       B bleu    3
   RGB(255,255,255), // WHITE    Up     Haut        W blanc   4
   RGB(255,255,0),   // YELLOW   Down   Bas         Y jaune   5
   RGB(127,127,127), // GRAY                          gris    6
   RGB(0,0,0)        // BLACK                         noir    7
};

static TCHAR colorsSet = TCHAR('1');         // default color set

static COLORREF UsedColors[8] = {
   defaultColors[0],defaultColors[1],defaultColors[2],
   defaultColors[3],defaultColors[4],defaultColors[5],
   defaultColors[6],defaultColors[7]
};

static CBrush BrushColors[8] = {
   UsedColors[0], UsedColors[1], UsedColors[2],
   UsedColors[3], UsedColors[4], UsedColors[5],
   UsedColors[6], UsedColors[7]
};

static bool colorsDecode(const CString &msg, COLORREF colors[6])
//=========================================================================//
{
   COLORREF tmp[6];
   BYTE i, j, n;

   if (msg.GetLength() < 36 || msg.Left(6) != "COLORS") return false;

   for (i = 0, n = 6; i < 6; i++) {
      static const CString decod("0 1 2 3 4 5 6 7 8 9 AaBbCcDdEeFf");
      COLORREF rgb = 0;
      int ix;

      for (j = 0; j < 6; j++)
         if ( (ix = decod.Find(msg.GetAt(n++))) < 0 ) return false;
         else rgb = (rgb << 4) | (ix >> 1);

      tmp[i] = rgb;
   }

   for (i = 0; i < 6; ++i) colors[i] = tmp[i];
   return true;
}

static CString colorsCode(const COLORREF colors[6])
//=========================================================================//
{
   CString msg;
   msg.Format("COLORS%06X%06X%06X%06X%06X%06X",
      colors[0],colors[1],colors[2],colors[3],colors[4],colors[5]);

   return msg;
}

TCHAR GetCharColor(BYTE color, bool full)
{
   static const CString tc("ORGBWY ?"), tcf("ORGBWY-*?");

   if (color > 8) color = 8;
   return (full)? tcf.GetAt(color): tc.GetAt(color);
}

//=========================================================================//
//                                                                         //
//                Compute the color from the number (0..5)                 //
//                                                                         //
COLORREF RGBDrawColor(BYTE color)         // number                        //
//=========================================================================//
{
   return color < 8? UsedColors[color]: UsedColors[6];
}

//=========================================================================//
//                                                                         //
//              Compute the text color from the background color           //
//                                                                         //
COLORREF RGBTextColor(BYTE color)         // background color              //
//=========================================================================//
{
   if (color < 8) {
      COLORREF rgb = UsedColors[color];
      int lum = Brightness(rgb);

      return (lum > 80)? RGB(0,0,0): RGB(255,255,255);
   } else {
      return RGB(0,0,0);
   }
}

//=========================================================================//
//                                                                         //
//                                                                         //
CBrush *GetBrush(BYTE color)
//=========================================================================//
{
   if (color < 8) return BrushColors + color;
   else return BrushColors + 6;
}

void SetBrush(BYTE color,              // number of color (0..5)           //
              COLORREF rgb)            // color                            //
//=========================================================================//
{
   if (color < 6) {
      if (BrushColors[color].m_hObject) BrushColors[color].DeleteObject();
      UsedColors[color] = rgb;
      BrushColors[color].CreateSolidBrush(rgb);
   }
}

void SetDefaultColors()
//=========================================================================//
{
   for (BYTE i = 0; i < 8; i++)
      SetBrush(i, UsedColors[i] = defaultColors[i]);

   colorsSet = TCHAR('1');
}

bool IsDefaultColors()
//=========================================================================//
{
   for (int i = 0; i < 6; i++) 
      if (UsedColors[i] != defaultColors[i]) return false;

   return true;
}

void ColorsExchange(BYTE u, BYTE v)
//=========================================================================//
{
   if (u != v) {
      COLORREF rgb_u = RGBDrawColor(u);
      SetBrush(u, RGBDrawColor(v));
      SetBrush(v, rgb_u);
   }
}

void ColorsExchange(const BYTE colors[6], bool old)
//=========================================================================//
{
   COLORREF oldColors[8];
   BYTE i, c;

   if (old)
      for (i = 0; i < 6; i++) oldColors[i] = UsedColors[i];
   else
      for (i = 0; i < 6; i++) oldColors[i] = defaultColors[i];

   oldColors[6] = defaultColors[6];
   oldColors[7] = defaultColors[7];

   for (i = 0; i < 6; i++) {
      if ((c = colors[i]) > 7) c = 6;
      SetBrush(i, oldColors[c]);
   }
}

//=========================================================================//
CString ColorsExchange(const CString &msg, const BYTE colors[6], bool invers)
//=========================================================================//
{
   COLORREF tmp0[6], tmp1[6];
   CString outMsg;
   BYTE i;

   if (!colorsDecode(msg, tmp0))
      return "";

   if (colors) {
      if (invers) for (i = 0; i < 6; ++i) tmp1[i] = tmp0[colors[i]];
      else        for (i = 0; i < 6; ++i) tmp1[colors[i]] = tmp0[i];
   }

   return colorsCode(tmp1) + msg.Mid(36);
}

//=========================================================================//
//                                                                         //
//                    Reading the colors in the register                   //
//                                                                         //
void InitColors()                                                          //
//=========================================================================//
{ 
   CString regColors = ::AfxGetApp()->GetProfileString("Settings", "ColorsRC");

   SetDefaultColors();
   ReadColors(regColors, true);
}

//=========================================================================//
//                                                                         //
//                 Memorization of the colors in the register              //
//                                                                         //
void SaveRegColors()                                                       //
//=========================================================================//
{ 
   ::AfxGetApp()->WriteProfileString("Settings", "ColorsRC", WriteColors());
}

//=========================================================================//
//                                                                         //
//                  Update the colors table from a string                  //
//                                                                         //
BOOL ReadColors(const CString &msg, bool update)                          //
//=========================================================================//
{
   COLORREF colors[6];
   BYTE i;

   if (!colorsDecode(msg, colors)) return FALSE;

   if (update) {
      for (i = 0; i < 6; i++) SetBrush(i, colors[i]);
      colorsSet = (msg.GetLength() > 36)? msg.GetAt(36): TCHAR(' ');
      return TRUE;
   } else {
      // return TRUE : new colors are present but different
      for (i = 0; i < 6; i++) if (colors[i] != UsedColors[i]) return TRUE;
      return FALSE;
   }
}

//=========================================================================//
//                                                                         //
BOOL CheckColorsSet(const CString &msg)                                    //
//=========================================================================//
{
   TCHAR set = (msg.GetLength() > 36)? msg.GetAt(36): TCHAR(' ');
   return set == colorsSet;
}

//=========================================================================//
//                                                                         //
t_COLORS oppositeColors(t_COLORS colors)
{
   t_COLORS out = 0;

   colors &= 0x3f;
   for (BYTE color = 0; colors; ++color, colors >>= 1)
      if (colors & 1) out |= 1 << (color ^ 1);

   return out;
}

//=========================================================================//
//                                                                         //
//        Output of the table of the 6 colors coded in hexadecimal         //
//                                                                         //
CString WriteColors()                                                      //
//=========================================================================//
{
   return colorsCode(UsedColors) + colorsSet;
}

//=========================================================================//
//
void PopulateColorsMenu(HMENU hMenu, UINT id, t_COLORS colors, BYTE flags)
//=========================================================================//
{
   if (colors) {
      // Built the menu with different colors entries
      for (BYTE i = 0; i < 6; i++) {

         if (colors & (1 << i))
            ::AppendMenu( hMenu, MF_OWNERDRAW, id + i, (LPCTSTR)i );
      }
   } else {

      ::AppendMenu(hMenu, MF_STRING, id + 6, "Effacer couleur");
   }
   if (flags & 1)
      ::AppendMenu(hMenu, MF_STRING, id + 7, "Effacer couleurs lies");
   if (flags & 2)
      ::AppendMenu(hMenu, MF_STRING, id + 8, "Copier  toute la face");
}

/////////////////////////////////////////////////////////////////////////////