// Colors.h: interface for the Colors utilities.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_COLORS_H__B66CC5DE_F362_4431_B0EC_0CEA20A5140D__INCLUDED_)
#define AFX_COLORS_H__B66CC5DE_F362_4431_B0EC_0CEA20A5140D__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

/////////////////////////////////////////////////////////////////////////////
// colors et brush

extern COLORREF RGBDrawColor(BYTE color);
extern COLORREF RGBTextColor(BYTE color);
extern TCHAR GetCharColor(BYTE color, bool full = false);

extern CBrush *GetBrush(BYTE color);
extern void SetBrush(BYTE color, COLORREF rgb);

extern void ColorsExchange(BYTE u, BYTE v);
extern void ColorsExchange(const BYTE colors[6], bool old = false);
extern CString ColorsExchange(const CString &msg, const BYTE colors[6], bool invers = false);
extern void SetDefaultColors();
extern bool IsDefaultColors();
extern void InitColors();
extern void SaveRegColors();

extern BOOL ReadColors(const CString &msg, bool update = false);
extern BOOL CheckColorsSet(const CString &msg);

extern CString WriteColors();

inline BYTE Brightness(COLORREF rgb) { return (BYTE)WORD((GetRValue(rgb) * 307
            + GetGValue(rgb) * 604 + GetBValue(rgb) * 113)>> 10); }

//*************** Correlation between colors and faces ********************//
// Give the default color corresponding to a facet number                  //
inline BYTE defColor(BYTE no)       // W Y R O B G                         //
   { static const BYTE cv[] = {4,5,1,0,3,2}; return no < 6? cv[no]: 0xFF; }
//
inline BYTE defNoColor(BYTE noSide) { return defColor(noSide / 9); }
//
inline BYTE oppositeColor(BYTE color) { return color ^ 1; }

//*************************************************************************//

typedef BYTE t_COLORS;  // signature color: each bit represents a color
                        // ie : bit0=> color 0, bit1=> color 1,...

extern t_COLORS oppositeColors(t_COLORS colors);

inline t_COLORS cvrtColor(BYTE color)
   { return (color < 6)? (1 << color): 0; }

inline BYTE colorCvrt(t_COLORS cvrt)
   { cvrt &= 0x3f; BYTE c = 0xff; while (cvrt) ++c, cvrt >>= 1; return c; }

// the signature color of a cubie is the 'or' (|) of the signatures of its
// visible faces. Each cubie of a Rubik's Cube has a different signature.

// count the colors in a signature color of a cube
inline BYTE colorsCount(t_COLORS cvrt)
   { cvrt &= 0x3f; BYTE c = 0; while(cvrt) c += cvrt&1, cvrt>>= 1; return c; }

// the center cubie has zero color, the cubies, centers sides, have one
// color, the edge cubies have two colors and the corner cubies have
// three colors

// for debug, visualize colors in a signature (in ascii)
inline CString outColors(t_COLORS cvrt) { CString out('-',6);
   for (BYTE i=0; i<6; i++, cvrt>>=1) if (cvrt&1) out.SetAt(i, GetCharColor(i));
   return out; }

extern void PopulateColorsMenu(HMENU hMenu, UINT id, t_COLORS colors, BYTE flags);

/////////////////////////////////////////////////////////////////////////////

#endif // !defined(AFX_COLORS_H__B66CC5DE_F362_4431_B0EC_0CEA20A5140D__INCLUDED_)
