// Completion.h: interface for the CCompletion class.
//
/////////////////////////////////////////////////////////////////////////////

#if !defined(AFX_COMPLETION_H__E07DB00A_8461_4DDC_B073_5B44AEAB14FC__INCLUDED_)
#define AFX_COMPLETION_H__E07DB00A_8461_4DDC_B073_5B44AEAB14FC__INCLUDED_

#include "CubeUtilities.h"

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

/////////////////////////////////////////////////////////////////////////////
// Determine the different possible configurations of the Rubik's Cube from
// the current state
/////////////////////////////////////////////////////////////////////////////

class CCompletion  
{
public:
   // construction / destruction
   CCompletion();
   virtual ~CCompletion();


   void Get(const BYTE facets[54]);

   inline void SetOptions(BYTE options)
      {  m_edges.SetOptions(options);
         m_corners.SetOptions(options); }

   // put color on a facet or delete one or more facets
   void SetAt(BYTE facet, BYTE color, BYTE facets[54], bool upd);
   void DelLinked(BYTE facet, BYTE facets[54]);
   BYTE FillFullSide(BYTE facet, BYTE facets[54], bool upd);

   // definition of possible colors for an update
   BYTE GetPossibleColors(BYTE noSide, const BYTE facets[54]) const;
   CString GetPossibleMsg(BYTE noSide, const BYTE facets[54]) const;

   // fills the faces that have only one possibility of color
   void Update(BYTE facets[54]);

   // facets center of faces (Rubik's Cube orientation)
   t_COLORS CubeOrientation(BYTE noSide, const BYTE facets[54]) const;
   int CentersUpdate(BYTE facets[54])const;

   CString getErrorMsg() const;

private:

   bool m_updOld;

   t_GROUP CornerUpdate(t_GROUP cornerColors, BYTE noCube);
   t_COLORS OrientationColors(BYTE side, const BYTE facets[54]) const;

   inline CString CenterSideMsg() const
      { return m_edges.m_colorsCoding? "":
            "Centres des faces  renseigner !\r\n"; }

   class CSlv
   {     // edges and corners are calculated separately by the same class

      CSlv(BYTE nbc);         // construction of class (edges=2 or corners=3)
      virtual ~CSlv();

      BYTE m_nbc;             // number of colors per structure (2 or 3)
                              // the difference between edges and corners

      struct Node {           // solver tree node
         BYTE side;           // side 0..23, in work buffer to filling
         t_COLORS colors;     // colors to filling
      } m_solvTbl[24],*m_pEnd;

      BYTE m_work[24];        // working buffer of facets colors
      t_COLORS m_solutions[2][24];// solutions classified according parity

      int m_sCnt;             // number of solutions found or error code
      BYTE m_flags;           // permutations parity of solutions
      bool m_computed;        // the solutions are computed

      BYTE m_refColors[13];   // reference signatures of 12 edges or 8 corners
      t_COLORS m_refCvrt[64]; // table no cubies 0..26 acceded by colors signature

      BYTE m_reversSides[54]; // to convert a facet (0..53) to side (0..23)
      BYTE m_convSides[24];   // to convert a side (0..23) to facet (0..53)

      BYTE m_orgSides[24];    // to give a begining side of each cubies

      t_COLORS m_colorsId[6]; // to convert a output colors bits
      BYTE m_colorId[6];      // to convert a output color
      bool m_colorsCoding;    // true => the colors are trancoded

      t_CUBIES m_cubesBits;   // bits of cubies of the computed type

      bool m_testOrientation; // enable the test of orientation the cubies
      bool m_testPermutation; // enable the test of permutations parity


      // Give all facets of the same cubie that the first facet 'side '
      inline t_GROUP getSidesNo(BYTE side) const { return (m_nbc == 3)?
         GROUP(side, side+1, side+2): GROUP(side, side+1); }

      // Give all colors of the same cubie that the first facet 'side '
      inline t_GROUP getColorsNo(BYTE side) const { return (m_nbc == 3)?
         GROUP(m_work[side], m_work[side+1], m_work[side+2]):
         GROUP(m_work[side], m_work[side+1]); }

      // Give a colors signature of the cubie that the first facet 'side '
      inline t_COLORS getColors(BYTE side) const { return (m_nbc == 3)?
         (1 << m_work[side])|(1 << m_work[side+1])|(1 << m_work[side+2]):
         (1 << m_work[side])|(1 << m_work[side+1]); }
   
      // Give an updated colors of the cubie with the facet 'side'
      inline t_GROUP fillColor(BYTE side, BYTE color) const
         {  BYTE orgs = m_orgSides[side];
            return GInsert(getColorsNo(orgs), color, side-orgs); }

      // Give the other edge side of the same cubie
      inline BYTE otherEdgeSide(BYTE side) const
         {  return (side == m_orgSides[side])? side + 1: side - 1; }

      // count of colors used in all edges or corners
      BYTE getColorsCount(BYTE color) const;

      // set the completly filled cubies
      t_CUBIES getFilledCubes() const;
      // bits color conversion from the internal representation to the
      // external representation
      t_COLORS colorsConvert(t_COLORS colors) const;

      t_COLORS getPossibleColors(BYTE side, BYTE flags) const;
      // provides for the possible colors for this side 0..23
      t_COLORS getFreeColors(BYTE side) const;
      // deduced the possible colors of the computed solutions
      t_COLORS getComputedColors(BYTE side, BYTE flags) const;

      t_COLORS checksCornerColor(BYTE side, t_COLORS cn) const;
      t_COLORS searchEdgeColors(t_CUBIES used) const;
      t_COLORS searchColors() const;

      // computs solutions
      int Computs();
      int InitNodeTbl();
      BYTE ValidityCheck() const;
      void copySolution(BYTE parity);

      int Create(const BYTE facets[54]);
      bool Compare(const BYTE facets[54]) const;
      BYTE Update(BYTE facets[54], BYTE flags);

      inline BYTE getFlags() const { return m_flags & 3; }

      CString getErrorString(int err, BYTE flags) const;

      inline CString getErrorString(BYTE flags = 3) const
         { return getErrorString(m_sCnt, flags); }

      inline int getStatus() const { return m_sCnt; }
      int getStatus(BYTE flags) const;

      inline void SetOptions(BYTE options)
         {  m_testOrientation = (options & 1)!= 0;
            m_testPermutation = (options & 2)!= 0; }

      CString Dump(const BYTE wrk[24]) const;

      friend class CCompletion;
   } m_edges, m_corners;
};

#endif // !defined(AFX_COMPLETION_H__E07DB00A_8461_4DDC_B073_5B44AEAB14FC__INCLUDED_)
