// CubeUtilities.cpp: implementation of the CubeUtilities utilities.
//
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <Shlwapi.h>
#include "CubeUtilities.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

short timeRel = -1;

//=========================================================================//
//                                                                         //
//            Convert an ASCII string into a table of 54 colors            //
//                                                                         //
int SetSideText(const CString &input,  // ASCII string input               //
                BYTE facets[54],       // output table 54 facets colors    //
               const CString &colorsId)// 6 color letters + 6 color faces  //
                                       // + 3 center face rotation letters //
//=========================================================================//
{
   CString codesF = colorsId.Mid(6, 6), codesC = colorsId.Left(6), codesO;
   int length = input.GetLength(), len = length - 10, i, j, n, k, face, color, r;
   BYTE bf[54], or[6];

   switch (colorsId.GetLength()) {
   case 12 : break;
   case 13 : codesO = "+#-"; break;
   case 15 : codesO = colorsId.Right(3); break;
   default : ASSERT(false); }

   // init the working buffers
   memset(bf, ' ', 54);
   memset(or, 0, 6);

   // decode the input string
   for (n = i = 0; i < len ; i++) {
      // search beginning of a side
      if ((face = codesF.Find(input[i])) < 0 || face >= 6
         || input[++i] != TCHAR(':')) continue;

      // copy the 9 facet colors of the side
      for (j = face * 9, k = j + 9; j < k; j++) bf[j] = (BYTE)input[++i];

      // extract the center side orientation char
      if (i + 1 < length && (r = codesO.Find(input[i + 1]))>= 0)
         or[face] = BYTE(r + 1), ++i;
   }

   // colors decoding
   for ( k = i = 0;;) {
      if ((color = codesC.Find(bf[i]))>= 0 && color < 6) {
         facets[i] = (BYTE)color;
         if (++i == 54) {
            // OK, end of buffer, add the orientation to center of face
            for (face = 0; face < 6; ++face)
               facets[face * 9 + 4] |= or[face] << 4;
            return 0;
         }
      } else if (k == 0) {
         // error of colors decoding: the color codes are also in center of
         // the sides
         codesC.Empty();
         for (j = 4; j < 54; j += 9)
            if (bf[j] > (BYTE)' ') codesC += (TCHAR)bf[j];
            else return -5;         // error: face not define
         // initializes a new colors decoding attempt, the last
         i = 0, k = 1;
      } else {
         return -1;                 // error: bad facet color
      }
   }
}

//=========================================================================//
//                                                                         //
//          Trace the faces of the Rubik's Cube in developed flat          //
//                   from a table of 54 colors of facets                   //
//                                                                         //
void DrawRCSides(CDC *pDC,                // drawing DC                    //
                 const CRect &rect,       // rectangle for the drawing     //
                 const BYTE facets[54],   // facet colors table            //
                 const CFont *pFont)
//=========================================================================//
{
   CRect rectSides[54];

   // calculate the rectangles of the 54 facets
   ComputRCSides(rect, rectSides);

   for (int n = 0; n < 54; n++) {
      BYTE color = facets[n] & 0x0f;
      CBrush *pBrushOld = pDC->SelectObject(GetBrush(color));
      CRect &r = rectSides[n];

      pDC->Rectangle(r);               // draw this rectangles
      if (pFont && n % 9 == 4) {
         pDC->SetTextColor(RGBTextColor(color));
         pDC->DrawText(CentersString(n / 9, facets[n] >> 4, "UDLRFB"),
                                      r, DT_CENTER|DT_SINGLELINE|DT_VCENTER);
      }
      pDC->SelectObject(pBrushOld);
   }
}

//=========================================================================//
//                                                                         //
//  Calculate the squares of the sides of the Rubik's Cube developed flat  //
//                                                                         //
void ComputRCSides(const CRect &rect,  // drawing space available          //
                   CRect rectSides[54])// faces rectangles table           //
//=========================================================================//
{  //      0-U
   // 2-L  4-F  3-R  5-B
   //      1-D          faces order :  U     D     L     R     F     B
   static const POINT positions[6] = {{1,0},{1,2},{0,1},{2,1},{1,1},{3,1}};
   int h = (rect.Height() - 10) / 9, e = (rect.Height() - (9 * h) - 8)/ 2;
   int orgx = (rect.left + rect.right - 12 * h - 3 * e)/ 2;
   int i, n, x, y;

   // calculate the rectangle faces
   for (i = 0, n = 0; i < 6; i++) {
      const POINT &org = positions[i];
      int ss = h * 3 + e;
      int ox = orgx + org.x * ss, oy = rect.top + 5 + org.y * ss;

      for (y = 0; y < 3; y++) {
         for (x = 0; x < 3; x++, n++) {
            int x0 = ox + x * h, y0 = oy + y * h;

            rectSides[n].SetRect(x0, y0, x0 + h - 1, y0 + h - 1);
         }
      }
   }
}

//=========================================================================//
//                                                                         //
//               Place the colors corresponding to the facets              //
//                                                                         //
void SetSideColors(t_GROUP noSides,    // facets involved                  //
                   t_GROUP noColors,   // corresponding colors             //
                   BYTE facets[54])    // table of 54 colors to update     //
//=========================================================================//
{
   BYTE noSide;

   while ((noSide = BYTE(noSides)) < 54) {
      facets[noSide] = BYTE(noColors);
      noSides = 0xFF000000|(noSides >> 8);
      noColors >>= 8;
   }
}

//=========================================================================//
//                                                                         //
//           Determines the colors corresponding to the facets             //
//                                                                         //
t_GROUP GetSideColors(t_GROUP noSides,       // facets involved            //
                   const BYTE facets[54])    // table of 54 colors         //
//=========================================================================//
{
   union { t_GROUP dw; BYTE b[4]; } data;
   BYTE noSide, color;

   data.dw = noSides;
   for (int i = 0; i < 4 && (noSide = data.b[i]) < 54; i++)
      color = facets[noSide] & 0x0F, data.b[i] = (color < 6)? color: 0xff;

   return data.dw;                           // output colors
}

/////////////////////////////////////////////////////////////////////////////
enum Facelet {       // Facelet indices
//  0  1  2  3  4  5  6  7  8   9 10 11 12 13 14 15 16 17 
   u0,u1,u2,u3,u4,u5,u6,u7,u8, d0,d1,d2,d3,d4,d5,d6,d7,d8,  // Up / Down
// 18 19 20 21 22 23 24 25 26  27 28 29 30 31 32 33 34 35
   l0,l1,l2,l3,l4,l5,l6,l7,l8, r0,r1,r2,r3,r4,r5,r6,r7,r8,  // Left / Right
// 36 37 38 39 40 41 42 43 44  45 46 47 48 49 50 51 52 53
   f0,f1,f2,f3,f4,f5,f6,f7,f8, b0,b1,b2,b3,b4,b5,b6,b7,b8   // Front / Back
};

// order of faces:   0=Up,   1=Down,  2=Left, 3=Right, 4=Front, 5=Back
// colors of faces:  White=4,Yellow=5,Red=1,  Orange=0,Blue=3,  Green=2

//        Up                        couleurs   W           W = U, Y = D,
//  Left  Front  Right  Back        faces    R B O G       R = L, O = R,
//        Down                      par defaut Y           B = F, G = B
//
//              u0 u1 u2              c26                 no Cubes  24 25 26
//     c21      u3 u4 u5       c23                          couche  21  U 23
//              u6 u7 u8                                superieure  18 19 20
//          c18   c19    c20      
//  l0 l1 l2    f0 f1 f2     r0 r1 r2     b0 b1 b2          couche  15  B 17
//  l3 l4 l5 c9 f3 f4 f5 c11 r3 r4 r5 c17 b3 b4 b5         mediane   L 13  R
//  l6 l7 l8    f6 f7 f8     r6 r7 r8     b6 b7 b8                   9  F 11
//           c0    c1    c2
//              d0 d1 d2                                    couche   6  7  8
//     c3       d3 d4 d5        c5                      inferieure   3  D  5
//              d6 d7 d8              c8                             0  1  2

static const t_GROUP sidesTbl[27] = {
   GROUP(d0, l8, f6),   // DLF cube 0
   GROUP(d1, f7),       // DF  cube 1
   GROUP(d2, f8, r6),   // DFR cube 2
   GROUP(d3, l7),       // DL  cube 3
   GROUP(d4),           // D   cube 4
   GROUP(d5, r7),       // DR  cube 5
   GROUP(d6, b8, l6),   // DBL cube 6
   GROUP(d7, b7),       // DB  cube 7
   GROUP(d8, r8, b6),   // DRB cube 8
   GROUP(l5, f3),       // LF  cube 9
   GROUP(f4),           // F   cube 10
   GROUP(r3, f5),       // RF  cube 11
   GROUP(l4),           // L   cube 12
   GROUP_EMPTY,         //     cube 13  central cube has no face
   GROUP(r4),           // R   cube 14
   GROUP(l3, b5),       // LB  cube 15
   GROUP(b4),           // B   cube 16
   GROUP(r5, b3),       // RB  cube 17
   GROUP(u6, f0, l2),   // UFL cube 18
   GROUP(u7, f1),		   // UF  cube 19
   GROUP(u8, r0, f2),   // URF cube 20
   GROUP(u3, l1),       // UL  cube 21
   GROUP(u4),           // U   cube 22
   GROUP(u5, r1),       // UR  cube 23
   GROUP(u0, l0, b2),   // ULB cube 24
   GROUP(u1, b1),       // UB  cube 25
   GROUP(u2, b0, r2) }; // UBR cube 26

t_GROUP getNoSides(BYTE cube)
   { ASSERT(cube < 27); return sidesTbl[cube]; }

// Tables and creation programs

// tables of the signature color for the cubes
static BYTE refCvrt[64] = {0, };

static bool makeRefCvrt()
{
   // provides the translation between the signature color and the number of
   // the cube
   BYTE cube, n1 = 0, n2 = 0, n3 = 0;

   memset(refCvrt, 0xFF, 64);

   // for all cubies of the Rubik's Cube
   for (cube = 0; cube < 27; cube++) {
      t_GROUP noSides = sidesTbl[cube];
      t_COLORS cvrt = cvrtColors( RefColors(noSides) );

      ASSERT(refCvrt[cvrt] == 0xFF);
      refCvrt[cvrt] = cube;               // table no cubies 0..26
   }

   return true;
}

// conversion from number face to number cube

static BYTE cvNoSide[54] = {0, };

static bool makeCvNoSide()
{
   memset(cvNoSide, 0xff, 54);

   for (BYTE noCube = 0; noCube < 27; ++noCube) {
      t_GROUP dw = sidesTbl[noCube];
      BYTE b;

      while ((b = BYTE(dw)) < 54)
         cvNoSide[b] = noCube, dw >>= 8;
   }

   for (BYTE f = 0; f < 54; ++f) ASSERT(cvNoSide[f] < 54);

   return true;
}

// cubies bits ranked by number of facets

static t_CUBIES cubesTbl[4] = {0,0,0,0};

static bool makeCubesTbl()
{
   for (BYTE noCube = 0; noCube < 27; noCube++)
      cubesTbl[SidesCount(sidesTbl[noCube])] |= cvrtCubie(noCube);

   return true;
}

/////////////////////////////////////////////////////////////////////////////

//=========================================================================//
//                                                                         //
BYTE GetCenterColor(t_GROUP cubeColors,// colors of the 3 faces Guidelines //
                    BYTE face)         // face to search a color           //
//=========================================================================//
{  //  colors : 0 1 2 3 4 5    faces : R,  L,  B,  F,  U,  D
   static const BYTE facetTable[6] = { 9,  8,  17, 16, 0,  1 };
   ASSERT(face < 6);
   BYTE fx = facetTable[face];

   return BYTE((cubeColors >> (fx & 0x18)) ^ (fx & 1));
}

//=========================================================================//
//                                                                         //
//         Defines the number of a cube from the signature color           //
//                      of this visible faces                              //
//                                                                         //
BYTE GetNoCube(t_COLORS cvrt)                                              //
//=========================================================================//
{
   if (refCvrt[0] == 0) makeRefCvrt();
   return cvrt < 0x40? refCvrt[cvrt]: 0xff;
}

//=========================================================================//
//                                                                         //
//           Convertion du no de face en no de cube appartenant            //
//                                                                         //
BYTE side2NoCube(BYTE noSide)                                              //
//=========================================================================//
{
   if (cvNoSide[0] == 0) makeCvNoSide();
   return noSide < 54? cvNoSide[noSide]: 0xff;
}

//=========================================================================//
//                                                                         //
//                    Gives the cubies completely filled                   //
//                                                                         //
t_CUBIES GetCompleteCubies(const BYTE facets[54])                            //
//=========================================================================//
{
   t_CUBIES cubes = CENTER_CUBIE;

   if (cvNoSide[0] == 0) makeCvNoSide();

   for (BYTE f = 0; f < 54; ++f) {
      if (facets[f] >= 6) cubes |= 1 << cvNoSide[f];
   }
   return cubes ^ ALL_CUBIES;
}

//=========================================================================//
//                                                                         //
//            List the numbers of visible faces of a givend cube           //
//                         (0 to 3 faces visible)                          //
//                                                                         //
t_GROUP GetNoSides(BYTE noCube)                                            //
//=========================================================================//
{
   return noCube < 27? sidesTbl[noCube]: GROUP_EMPTY;
}

//=========================================================================//
//                                                                         //
//           Gives the default color of a list of visible faces            //
//                                                                         //
t_GROUP RefColors(t_GROUP noSides)                                         //
//=========================================================================//
{
   union { t_GROUP dw; BYTE b[4]; } data;

   data.dw = noSides;
   for (int i = 0; i < 4; i++)
      data.b[i] = defNoColor(data.b[i]);

   return data.dw;
}

//=========================================================================//
//                                                                         //
//       Extract from the list the two central facets filled in            //
//       allowing to calculate the orientation of a Rubik's Cube.          //
//            Do not verify that other facets can indicate                 //
//                        a different direction.                           //
//                                                                         //
t_GROUP GetOrientSides(const BYTE facets[54])                              //
//=========================================================================//
{
   // analyze aspects of face centers
   for (UINT n = 4, m; n < 54; n += 9) {
      if ((facets[n] & 7)== 7) continue;
      // jumps (possibly) the facet opposite to the first
      m = ((n / 9)| 1) * 9 + 4;
      // search the next
      while ((m += 9) < 54) {
         if ((facets[m] & 7) < 6)
            // second facet found
            return GROUP(n, m);
      }
      break;                        // exit, second facet not found
   }
   return GROUP_EMPTY;              // not found (first or second)
}

/////////////////////////////////////////////////////////////////////////////
// The definiton string of the Identity cube in Singmaster notation is
static const BYTE SingmasterEdgeToFacelet[24] = {
   // edges UF UR UB UL DF DR DB DL FR FL BR BL
   u7,f1,u5,r1,u1,b1,u3,l1,d1,f7,d5,r7,d7,b7,d3,l7,f5,r3,f3,l5,b3,r5,b5,l3};
static const BYTE SingmasterCornerToFacelet[24] = {
   // corners UFR URB UBL ULF DRF DFL DLB DBR
	u8,f2,r0,u2,r2,b0,u0,b2,l0,u6,l2,f0,d2,r6,f8,d0,f6,l8,d6,l6,b8,d8,b6,r8};
//=========================================================================//
//                                                                         //
CString ToSingmasterString(const BYTE facets[54],  // faces order UDLRFB   //
         const CString &colorsId)      // 6 letters of colors + 6 of faces //
//=========================================================================//
{
   CString string(TCHAR(' '), 67);
   TCHAR colors[6];
   int i, j;

   // get colors of the center of the faces
   for (i = 0; i < 6; i++)
      colors[facets[i * 9 + 4]& 7] = colorsId[i + 6];

   for (i = j = 0; i < 24; i++) {
	   string.SetAt(j++, colors[facets[SingmasterEdgeToFacelet[i]]& 7]);
	   // add spaces if necessary
      if (i % 2 == 1) j++;
   }

   for (i = 0; i < 24; i++) {
	   string.SetAt(j++, colors[facets[SingmasterCornerToFacelet[i]]& 7]);
	   // add spaces if necessary
	   if ( i % 3 == 2 ) j++;
   }

   return string;
}

//=========================================================================//
//                                                                         //
bool SingmasterToFacets(const CString &msg, BYTE facets[54])               //
//=========================================================================//
{
   static const BYTE colors[6] = {4,5,1,0,3,2}; // Singmaster colors convert
   static const CString sgm("UDLRFB(/\n\r");    // Singmaster faces names
   int i, j, ix;

   for (i = 0; i < 54; i++) facets[i] = 0xFF;

   // set center of faces colors
   for (ix = 0; ix < 6; ix++) facets[ix * 9 + 4] = colors[ix];

   // convert Singmaster message
   for (i = 0, j = 0; i < msg.GetLength(); i++) {
      if ((ix = sgm.Find(msg.GetAt(i))) < 0) continue;
      else if (ix >= 6) break;

      if (j < 24)      facets[SingmasterEdgeToFacelet[j]] = colors[ix];
      else if (j < 48) facets[SingmasterCornerToFacelet[j-24]] = colors[ix];
      j++;
   }

   return (j == 48);
}

//=========================================================================//
//                                                                         //
//           Test whether the colors of this cubie are legal               //
//                                                                         //
char CheckSidesColors(t_GROUP colors,  // colors has identified            //
                      BYTE *pNoCube)   // cube corresponding to the colors //
//=========================================================================//
{
   t_COLORS cvrt = cvrtColors(colors);
   BYTE noCube = GetNoCube(cvrt);   // cubie corresponding to this signature
   BYTE nbc = colorsCount(cvrt);

   if (nbc != SidesCount(colors) || noCube >= 27)
      return -9;     // illegal combination of colors

   if (pNoCube) *pNoCube = noCube;  // cubie number ouput

   t_GROUP refColors = RefColors(sidesTbl[noCube]);
   ASSERT(cvrt == cvrtColors(refColors));

   // Research the number in rotations required to get the color group
   // corresponding to this cubie
   for (char n = 0; n < 3; n++, colors = Rotate(colors))
      if (colors == refColors)
         return n;   // number in rotations required (0,1 ou 0,1,2)

   // for a edge cube in 2 colors there are only two possible combinations
   // and legal, for a corner cube with 3 colors there are 6 possible
   // combinations of which only 3 are legal. The program also works for a
   // center face facet with one color !

   return -10;       // legal combination of colors, but disordered
}

//=========================================================================//
//                                                                         //
//        Calculate the number of permutations in places of cubies         //
//                                                                         //
BYTE PermutationsParity(const BYTE cubes[],  // cubies no table            //
                        BYTE len)            // table size                 //
//=========================================================================//
{
   BYTE parity = 0;
   // Permutation parity can be computed by counting the number of reversals
   // in the permutation sequence, - i.e., the number of pairs (p,q) such
   // that p>q and p precedes q. Then determine if the result is even or odd.
   // Do this for both edges (EPP) and corners (CPP). A configuration is
   // reachable if EPP=CPP or EPP+CPP is odd. (August/September cube.lovers -
   // Vanderschel/Saxe)
   for (BYTE p = 1; p < len; p++)
      for (BYTE q = p, k = cubes[p - 1]; q < len; q++)
         if (k > cubes[q]) parity ^= 1;

   return parity;
}

//=========================================================================//
//                                                                         //
//          Fournit la signature couleurs d'un groupe de facettes          //
//                                                                         //
t_COLORS GetColors(t_GROUP noSides, const BYTE facets[54])                 //
//=========================================================================//
{
   BYTE side, color, colors = 0;

   while ((side = BYTE(noSides)) < 54) {
      noSides >>= 8;
      if ((color = facets[side] & 7) < 6) colors |= 1 << color;
      else colors += 0x40;          // compte les facettes non renseignees
   }
   return colors;
}

//=========================================================================//
//                                                                         //
//               Complete les premieres couleurs manquantes                //
//                                                                         //
t_GROUP FillColors(t_GROUP colors, BYTE c1, BYTE c2)                        //
//=========================================================================//
{
   union { t_GROUP gr; BYTE b[4]; } c;

   c.gr = colors;
   for (int i = 0; i < 3; i++)
      if (c.b[i] == 0xff) c.b[i] = c1, c1 = c2, c2 = 0xff;

   return c.gr;
}

//=========================================================================//
//                                                                         //
//             Change the color defined by the facet number                //
//                                                                         //
t_GROUP FillColor(BYTE side, BYTE color, const BYTE facets[54])            //
//=========================================================================//
{
   union { t_GROUP gr; BYTE b[4]; } colors, nSides;
   int i;

   // extract all visible facets of the cube for the designated facet
   nSides.gr = GetAllSides(side);
   // the facets colors
   colors.gr = GetSideColors(nSides.gr, facets);
   // search the facet in all facets of the cube
   for (i = 0; nSides.b[i] != side; i++) ASSERT(i < 2);
   // change the color
   colors.b[i] = color;
   // give all the colors of the cube changed
   return colors.gr;
}

//=========================================================================//
//                                                                         //
//      Provide all cubies having the same number of visible facets        //
//                                                                         //
t_CUBIES GetCubiesVSC(BYTE nbc)                                            //
//=========================================================================//
{
   if (cubesTbl[0] == 0) makeCubesTbl();

   return cubesTbl[nbc & 3];
}

/////////////////// Rotation of the facets side center //////////////////////

//=========================================================================//
//                                                                         //
//         Tests the state of rotation of the facets center side           //
//                         of the Rubik's Cube                             //
//           This test is only meaningful for a ordered Cube               //                                                              //
//                                                                         //
bool IsCentersLegal(DWORD centers)                                         //
//=========================================================================//
{
   DWORD counts = 0;

   for (centers &= 0x00FFFFFF; centers; centers >>= 4)
      counts += centers;

   return (counts & 1) == 0;
}

//=========================================================================//
static const TCHAR *orient[4] = { NULL, "+", "++", "-" }; // rotations
static const CString codeCf("RLBFUD"); // names for the facets center of side

inline static int FindFace(int face)
   { int f = codeCf.Find(TCHAR(face)); return f < 0? face: f; }

static CString GetMsgCenters(DWORD centers, int face, bool all, bool space)
{
   CString msg;

   ASSERT(face >= 0 && face < 6);

   BYTE or = BYTE(centers >> (face * 4)) & 3;
   if (or || all) msg = CentersString(face, or, "RLBFUD");

   if (space && msg != "") msg.Insert(0, TCHAR(' '));

   return msg;
}

//=========================================================================//
//                                                                         //
void PopulateOrientMenu(HMENU hMenu, UINT id,
                        BYTE face, BYTE faceColor,
                        const CString &faceLeters)
//=========================================================================//
{
   BYTE fo = (faceColor >= 0x40)? 0: faceColor >> 4;

   for (BYTE or = 0; or < 4; ++or) { // for the 4 orientations
      CString text = CentersString(face / 9, or, faceLeters);
      UINT flags = (fo == or)? MF_STRING|MF_ENABLED|MF_CHECKED:
                                    MF_STRING|MF_ENABLED;

      ::AppendMenu(hMenu, flags, id + or, text);
   }
}

//=========================================================================//
//                                                                         //
//       Convert the rotation of all center side of the Rubik's Cube       //
//                                                                         //
CString GetMsgCenter(DWORD centers,          // counts of centers          //
                             int face)       // face number or all         //
//=========================================================================//
{
   CString msg, tmp;

   face = FindFace(face);
   if (face >= 0 && face < 6) {

      msg = GetMsgCenters(centers, face, true, false);
   } else {

      for (int i = 0; i < 6; i++)
         msg += GetMsgCenters(centers, i, false, msg != "");
   }

   return msg;
}



//=========================================================================//
//                                                                         //
//       Calculate the rotation of one center side of the Rubik's Cube     //
//                                                                         //
int CenterFaceOrient(DWORD centers,          // counts of centers          //
                             int face)       // face number                //
//=========================================================================//
{
   face = FindFace(face);

   if (face >= 0 && face < 6)
      return (centers >> (face * 4)) & 3;

   return -7;
}

//=========================================================================//
//                                                                         //
//       Changes the rotation of one center side of the Rubik's Cube       //
//                                                                         //
DWORD SetCenterFaceOrient(DWORD centers,     // counts of centers          //
                            int face,        // face number                //
                                 int orient) // new count                  //
//=========================================================================//
{
   face = FindFace(face);

   if (face >= 0 && face < 6) {
      orient &= 3;
      centers &= ~( 0x0F  << (face * 4));
      centers |=  (orient << (face * 4));
   }

   return centers;
}

//=========================================================================//
//                                                                         //
//           Change the location of counts of the central facets           //
//                                                                         //
DWORD CenterTranslate(DWORD centers,         // facets center of face      //
                        const BYTE oper[6],  // permutation order          //
                             WORD sgn)       // sign change order          //
//=========================================================================//
{
   DWORD out = 0, t;

   for (int i = 0; i < 6; i++, centers >>= 4)
      t = 3 &((sgn >> i)& 1? 4 - centers: centers), out |= t << (oper[i] * 4);

   return out;
}

//=========================================================================//
//                                                                         //
//                    Read a ascii string from a file                      //
//                                                                         //
CString ReadText(const CString &fileName)                                  //
//=========================================================================//
{
   CFile file;
   if (!file.Open(fileName, CFile::modeRead))
      return "";

   int len = (int)file.GetLength();
   CString text(char(0), len);

   file.Read(text.GetBuffer(len), len);

   return text;
}

//=========================================================================//
//                                                                         //
//                  Add a library path to the file name                    //
//                                                                         //
CString FixePath(const CString &fileName)                                  //
//=========================================================================//
{
   TCHAR path[MAX_PATH];

   // Computs file name of the solver tables
   ::GetModuleFileName( ::AfxGetInstanceHandle(), path, MAX_PATH );
   ::PathRemoveFileSpec(path);
   ::PathAppend(path, fileName);
   //
   return path;
}

//=========================================================================//
//                                                                         //
//          Error messages on the construction of a Rubik's Cube           //
//                                                                         //
CString GetErrorMessage(int er,              // error code                 //
                        const CString &cr)   // End of line                //
//=========================================================================//
{
   CString msg;

   if (er < 0) {
      static const TCHAR *pChar[] = {
         "",
         // erreurs de lecture d'une definition couleurs facettes
         "Format de dfinition des couleurs facettes invalide.",     // -1
         "Double dfinition des couleurs d'une face.",               // -2
         "Dcompte des couleurs facettes invalide."                  // -3
         "Couleur(s) facette(s) non renseigne(s).",                 // -4
         "Couleur(s) facette(s) non reconnue(s).",                   // -5
         "Identification de face du Rubik's Cube erronne.",         // -6
         "Aucune solution n'est trouve.",                           // -7
         // erreurs de construction d'un Rubik's Cube
         "Couleurs facettes centres de faces illgales.",            // -8
         "Couleurs d'un cube lmentaire illgales.",                // -9
         "Ordre de couleurs d'un cube lmentaire invalide.",        //-10
         "Trop de cubes sont indfinis.",                            //-11
         "Double combinaisons de couleurs de cubes lmentaires."    //-12
      };

      if (er >= -12) msg = pChar[-er] + cr;
      else msg.Format("erreur %d.%s", -er, cr);

   } else {
      // organizational mistakes of color facets
      if (er & 1)
         msg += "Les orientations des artes sont incompatibles." + cr;
      if (er & 2)
         msg += "Les orientations des sommets sont incompatibles." + cr;
      if (er & 4)
         msg += "Permutations d'emplacement des cubes est impaire." + cr;
      if (er & 8)
         msg += "L'orientation des facettes centrales est illgale." + cr;
   }

   if (msg.IsEmpty() && er != 0) msg.Format("erreur inconnue %02x.%s", er, cr);

   return msg;
}

/////////////////////////////////////////////////////////////////////////////