// CubeUtilities.h: interface for the CubeUtilities utilities.
//
/////////////////////////////////////////////////////////////////////////////

#if !defined(AFX_CUBEUTILITIES_H__E96304BC_CA28_4540_80BF_D1B5D9CD7D12__INCLUDED_)
#define AFX_CUBEUTILITIES_H__E96304BC_CA28_4540_80BF_D1B5D9CD7D12__INCLUDED_

#include "Colors.h"
#include "CmdString.h"

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


/////////////////////////////////////////////////////////////////////////////
   // BYTE scmd : compressed representation of a command
   //    bits 0,1 : axe + 1 : x => 1, y => 2, z => 3, 0=>null command,
   //    bits 2,3 : plan 0 => 0, 4 => 1, 8 => 2, 0xC => all plans,
   //    bits 4,5,6: provision
   //    bit  7   : sens rotation 0=> CW, 1=> CCW */

inline t_Mask Command(BYTE scmd)
   { return (scmd & 3)? Command(--scmd & 0x03, (scmd & 0x0C)>> 2, (scmd & 0x80)!= 0): 0; }

inline BYTE PlanMed(BYTE scmd)
   { return (scmd & 0xF3) ^ 0x04; }

//   { return (scmd & 0xF3) ^ ((scmd & 0x0C0)? 0x84: 0x04); }
inline BYTE PlanAll(BYTE scmd)
   { return scmd | 0x0C; }

//   { return (scmd & 0x0C)? (scmd | 0x0C) ^ 0x80: (scmd | 0x0C); }
inline BYTE ShortCmd(short axe, short plan, bool neg)
   { return BYTE(++axe & 3)|BYTE((plan & 3)<<2)|(neg? 0x80: 0); }
/////////////////////////////////////////////////////////////////////////////
// Gives the cubie number with that default colors signature
extern BYTE GetNoCube(t_COLORS cvrt);

// Gives the number of the cubie with this visible facet
extern BYTE side2NoCube(BYTE noSide);


/////////////////////////////////////////////////////////////////////////////
// conversion entre texte et table de 54 facettes


extern int SetSideText(const CString &input, BYTE facets[54],
                                 const CString &colorsId = "ORGBWYDRBULF");

// representation of the table of the facets in held at flat
extern void ComputRCSides(const CRect &rect, CRect rectSides[54]);
extern void DrawRCSides(CDC *pDC, const CRect &rect, const BYTE facets[54],
                        const CFont *pFont);
//
extern CString ToSingmasterString(const BYTE facets[54],
                                 const CString &colorsId = "      UDLRFB");
extern bool SingmasterToFacets(const CString &msg, BYTE facets[54]);

// descripteur de face 2D
typedef struct { t_Mask plan; char x, y; } t_Side;


/////////////////////////////////////////////////////////////////////////////
typedef DWORD t_CUBIES;  // each bit represents a cubie number

const t_CUBIES ALL_CUBIES = (1 << 27) - 1;
const t_CUBIES CENTER_CUBIE = 1 << 13;

inline t_CUBIES cvrtCubie(BYTE noCube)   // convert cubie number in bit
   { return (noCube < 27)? (1 << noCube): 0; }
inline BYTE cubieCvrt(t_CUBIES cubes)    // return the last cubie number
   { BYTE c = 0xff; while (cubes) c++, cubes >>= 1; return c; }
inline BYTE cubiesCount(t_CUBIES cubes)  // counts the cubies
   { BYTE c = 0; while (cubes) c += BYTE(cubes&1), cubes >>= 1; return c; }

// provide all cubes having the same number of visible facets
extern t_CUBIES GetCubiesVSC(BYTE nbc);

// gives the cubes completely filled
extern t_CUBIES GetCompleteCubies(const BYTE facets[54]);

/////////////////////////////////////////////////////////////////////////////
typedef DWORD t_GROUP;  // the first bytes contain the number of facets
                        // (0 .. 53) or colors (0 .. 5). The unused bytes
                        // are set to 0xff.

inline t_GROUP GROUP(BYTE a= 0xff, BYTE b= 0xff, BYTE c= 0xff, BYTE d= 0xff)
   { return a|(b<<8)|(c<<16)|(d<<24); }

inline BYTE BYTE_0(t_GROUP gr) { return BYTE(gr); }
inline BYTE BYTE_1(t_GROUP gr) { return BYTE(gr >> 8); }
inline BYTE BYTE_2(t_GROUP gr) { return BYTE(gr >> 16); }
inline BYTE BYTE_3(t_GROUP gr) { return BYTE(gr >> 24); }
inline BYTE BYTE_N(t_GROUP gr, BYTE n) { return BYTE(gr >> (n << 3)); }
inline t_GROUP GInsert(t_GROUP gr, BYTE color, BYTE n)
   { n <<= 3; return (gr & ~(0xff << n)) | (color << n); }

const t_GROUP GROUP_EMPTY = DWORD(-1);
// the three sides guidelines to provide Rubik's Cube direction
const t_GROUP CENTER_FACES = GROUP(4, 4 + 2*9, 4 + 4*9);

// Give all visible facets belonging to a cubie
extern t_GROUP getNoSides(BYTE noCube);

extern BYTE GetCenterColor(t_GROUP cubeColors, BYTE face);

// gives the opposite facet to said facet, center of face
inline BYTE oppositeSide(BYTE side)
   { return (side % 9 == 4)? ((side / 9) ^ 1) * 9 + 4: 0xff; }

// Count the components (colors, dimensions,...) contained in a t_GROUP
inline BYTE SidesCount(t_GROUP gr)
   { return ((BYTE_0(gr)!=0xff)? 1:0)+((BYTE_1(gr)!=0xff)? 1:0)
            +((BYTE_2(gr)!=0xff)? 1:0) +((BYTE_3(gr)!=0xff)? 1:0); }

// Compute the color signature corresponding to a color group
inline t_COLORS cvrtColors(t_GROUP co) { return cvrtColor(BYTE_0(co))|
      cvrtColor(BYTE_1(co))|cvrtColor(BYTE_2(co))|cvrtColor(BYTE_3(co)); }

// gives the type of the facet (number of visible facets per cubie)
inline BYTE getFacetType(BYTE side)   // return 1, 2 or 3
   { side %= 9; return (side == 4)? 1: 3 - (side & 1); }


// Calculate the axis normal to the face 0=> x, 1=> y, 2=> z
inline BYTE getRAxe(BYTE side) { return side < 54? side / 18: 0xFF; }

// Give default colors corresponding to the numbers of facets
extern t_GROUP RefColors(t_GROUP noSides);
// Give default colors signature corresponding to a cubie
inline t_COLORS GetRefColors(BYTE noCube)
   {  return cvrtColors(RefColors(getNoSides(noCube))); }

// Give all facets of the same cubie that the facet 'side '
inline t_GROUP GetAllSides(BYTE side)
   { return getNoSides(side2NoCube(side)); }

// Gives the colors corresponding to the numbers of facets
extern t_GROUP GetSideColors(t_GROUP noSides, const BYTE facets[54]);
// Put the defined colors in the facet table
extern void SetSideColors(t_GROUP noSides, t_GROUP noColors, BYTE facets[54]);


// Extract numbers of facets filled in, allowing the calculation
// of the orientation
extern t_GROUP GetOrientSides(const BYTE facets[54]);
// Gives the signature color corresponding to the group of facets
extern t_COLORS GetColors(t_GROUP noSides, const BYTE facets[54]);
                 //
// center faces colors
inline BYTE ExtractSidesColors(BYTE no, t_GROUP colors)
   { return (no & 1) ^ BYTE(colors >> ((no / 2)* 8)); }

// Switch the colors or facets in the variable t_GROUP, corresponding to
// rotations of a legal cubie for edge or corner.
inline t_GROUP Rotate(t_GROUP colors) { BYTE b;
   if ((b=BYTE_3(colors))!=0xff) return (colors<<8)|b;           // 4 couleurs
   if ((b=BYTE_2(colors))!=0xff) return (colors<<8)|b|0xff000000;// 3 couleurs
   if ((b=BYTE_1(colors))!=0xff) return (colors<<8)|b|0xffff0000;// 2 couleurs
   return colors; }                 // 1 ou 0 couleur : on ne fait rien

// Rotation of colors for the n positions (0 to 3)
inline t_GROUP Rotate(t_GROUP colors, char n) { ASSERT(n >= 0 && n < 4);
   while (n > 0) colors = Rotate(colors), --n; return colors; }

// Check the color group and counts the number of rotation from the origin
extern char CheckSidesColors(t_GROUP colors, BYTE *pNoCube = NULL);

// Calculates the parity of the number of permutations of cubie space
extern BYTE PermutationsParity(const BYTE cubes[], BYTE len);

// Complete the first missing color
extern t_GROUP FillColors(t_GROUP colors, BYTE c1, BYTE c2 = 0xff);
// Change the color defined by the facet number, provides the corresponding
// color group
extern t_GROUP FillColor(BYTE side, BYTE color, const BYTE facets[54]);

/////////////////////////////////////////////////////////////////////////////
// Rotations of the facets center of face


inline DWORD SetCenters(BYTE rR, BYTE rL, BYTE rB, BYTE rF, BYTE rU, BYTE rD)
   { return rR |(rL << 4)|(rB << 8)|(rF << 12)|(rU << 16)|(rD << 20); }
   // cnt: -1,1,2; face: 0=>R, 1=>L, 2=>B, 3=>F ,4=>U, 5=>D
inline DWORD SetCenter(char cnt, BYTE face)
   { ASSERT(face < 6); return (cnt & 3) << (face * 4); }

extern bool IsCentersLegal(DWORD center);

inline CString CentersString(BYTE face, BYTE or, const CString &faceLeters)
   {  static const TCHAR *orient[4] = { NULL, "+", "++", "-" }; // rotations
      CString text = faceLeters.Mid(face, 1); if (or &= 3) text += orient[or];
      return text; }

extern CString GetMsgCenter(DWORD centers, int face = -1);
extern int CenterFaceOrient(DWORD centers, int face);
extern DWORD SetCenterFaceOrient(DWORD centers, int face, int orient);

inline DWORD NegateCenters(DWORD centers)
      { return ( (0x00444444 - centers) & 0x00333333 ); }

inline bool IsChgSign(DWORD c) { return ( (c & 0x00111111) != 0 ); }

DWORD CenterTranslate(DWORD centers, const BYTE oper[6], WORD sgn = 0);

extern void PopulateOrientMenu(HMENU hMenu, UINT id, BYTE face,
                               BYTE faceColor, const CString &faceLeters);

/////////////////////////////////////////////////////////////////////////////

// DEBUG fonctions
// convertit la couleur en un char ASCII
inline CString color2Str(BYTE color) { static const CString c("ORGBWY");
   return color < 6? c.Mid(color, 1): CString('?'); }
inline CString cvrt2Str(t_COLORS cvrt) { CString msg;
   for (BYTE c = 0; c < 6; c++) if (cvrt & (1 << c)) msg += color2Str(c);
   return msg.IsEmpty()? "-": msg; }
// convertit un groupe de couleurs en chaine ASCII
inline CString colors2Str(t_GROUP colors) { return color2Str(BYTE(colors>>16)) +
   color2Str(BYTE(colors>>8)) + color2Str(BYTE(colors)); }


// conversion d'un ID resource string en sa chaine ASCII correspondante
inline CString String(UINT id)
   { CString msg; msg.LoadString(id); return msg; }

// lecture d'un fichier de texte
extern CString ReadText(const CString &fileName);
// ajoute le dossier contenant le fichier '.exe' au nom du fichier
extern CString FixePath(const CString &fileName);

/////////////////////////////////////////////////////////////////////////////

extern CString GetErrorMessage(int err, const CString &cr = "\n");

/////////////////////////////////////////////////////////////////////////////
#endif // !defined(AFX_CUBEUTILITIES_H__E96304BC_CA28_4540_80BF_D1B5D9CD7D12__INCLUDED_)
