//
//                     COMPUTS AND DRAW A RUBIK'S CUBE
//
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Cubes.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//=========================================================================//
//
//       |           |
//    ---2---+---+---0---
//       | 0 | 0 | 0 |
//       +---+---+---+                Arrows point to "top"
//       | 0 Right 0 |                    of the cube
//       +---+---+---+
//       | 0 | 0 | 0 |           |                  |y
//    ---3---+---+---1---+---+---0---               |_ _ _ _ _ _
//       | 5 | 5 | 5 | 2 | 2 | 2 |                  |\_4_\_4_\_4_\
//       +---+---+---+---+---+---+                  |1|\_4_\_4_\_4_\ 
//       | 5 Down| 5 | 2  Back 2 |                  | |1|\_4_\_4_\_4_\ 
//       +---+---+---+---+---+---+                  |1| |1| 3 | 3 | 3 |
//       | 5 | 5 | 5 x 2 | 2 | 2 |           |      | |1| |___|___|___|
//   ----7---+---+--z5y--+---+---4---+---+---0---   |1| |1| 3 | 3 | 3 |
//       |           z 1 | 1 | 1 | 4 | 4 | 4 |       \|1| |___|___|___|--x
//                   +---+---+---+---+---+---+         \|1| 3 | 3 | 3 |
//                   | 1 Left  1 | 4 |Up | 4 |           \|___|___|___|
//                   +---+---+---+---+---+---+             \
//                   | 1 | 1 | 1 | 4 | 4 | 4 |               \
//                ---7---+---+---6---+---+---2---             z
//                   |           | 3 | 3 | 3 |
//                   |           +---+---+---+
//                               | 3 Front 3 |
//                               +---+---+---+
//                               | 3 | 3 | 3 |
//                           ----7---+---+---3---
//                               |           |
//
//=========================================================================//
// face orientation (default color is indicative)
static const char m0[3][3] = {0,-1,0, 1,0,0, 0,0,1};  // R, orange
static const char m1[3][3] = {0,1,0, -1,0,0, 0,0,1};  // L, rouge

static const char m2[3][3] = {-1,0,0, 0,-1,0, 0,0,1}; // B, vert
static const char m3[3][3] = {1,0,0, 0,1,0, 0,0,1};   // F, bleu (unit)

static const char m4[3][3] = {1,0,0, 0,0,1, 0,-1,0};  // U, blanc
static const char m5[3][3] = {1,0,0, 0,0,-1, 0,1,0};  // D, jaune
                
// faces names (x=> R-L, y => B-F, z=> U-D
static const t_Side dscr[6] = {
//  plan  |orient,      position x,y
   {0x0600|mPack(m0), 5,3},    // 0 Right
   {0x8600|mPack(m1), 1,3},    // 1 Left
   {0x1800|mPack(m2), 7,3},    // 2 Back
   {0x9800|mPack(m3), 3,3},    // 3 Front
   {0x6000|mPack(m4), 3,1},    // 4 Up
   {0xE000|mPack(m5), 3,5}     // 5 Down
};

t_Matrix GetOrientationSide(BYTE noSide)
   { return t_Matrix(dscr[noSide].plan); }

// table of faces for a view held flat R, L, B, F, U, D to generate a belt
// to extract the ascii position
static const t_Side dscrText[6] = {    // band R L B F U D
   {0x0600|mPack(m0), 1,5},{0x8600|mPack(m1), 1,1},
   {0x1800|mPack(m2), 1,7},{0x9800|mPack(m3), 1,3},
   {0x6000|mPack(m4), 1,9},{0xE000|mPack(m5), 1,11}
};
static const t_Side dscrBand[6] = {    // band R L B F U D
   {0x0600|mPack(m0), 1,1},{0x8600|mPack(m1), 1,3},
   {0x1800|mPack(m2), 1,5},{0x9800|mPack(m3), 1,7},
   {0x6000|mPack(m4), 1,9},{0xE000|mPack(m5), 1,11}
};
static BYTE moveCubes[21] =
   {0,1,2,3,5,6,7,8,9,11,13,15,17,18,19,20,21,23,24,25,26};

//=========================================================================//


/////////////////////////////////////////////////////////////////////////////
//                             class CCubes                                //
/////////////////////////////////////////////////////////////////////////////

//=========================================================================//
//                                                                         //
CCubes::CCubes(const CCubes *pCubes,   // copy position of this Cube       //
         float dim,           // size of a Rubik's Cube face               //
         int *pYObs,          // distance viever                           //
         const CMatrix *pMro, // transition matrix repere Cube Viewer      //
         CMatrix *pMmr)       // transition matrix repere part Mobile Cube //
//=========================================================================//
{
   ASSERT(initMatrixTest());
   m_box.SetRectEmpty();
   m_dim  = dim;
   m_pYObs= pYObs;
   m_pMro = pMro;
   m_pMmr = pMmr;
   m_nbSide = 0;
   m_oldMove = pMmr? pMmr->Pack(): 0;

   if (pCubes) memcpy(m_cube, pCubes->m_cube, 54); else Init();
}

//=========================================================================//
//                                                                         //
CCubes::CCubes(float dim,     // size of a Rubik's Cube face               //
               int *pYObs,    // distance viever                           //
         const CMatrix *pMro, // transition matrix repere Cube Viewer      //
               CMatrix *pMmr) // transition matrix repere part Mobile Cube //
//=========================================================================//
{
   ASSERT(initMatrixTest());
   m_box.SetRectEmpty();
   m_dim  = dim;
   m_pYObs= pYObs;
   m_pMro = pMro;
   m_pMmr = pMmr;
   m_nbSide = 0;
   m_oldMove = pMmr? pMmr->Pack(): 0;

   Init();
}

//=========================================================================//
//                                                                         //
void CCubes::SetScreen(int cx, int cy)                                     //
//=========================================================================//
{
   m_screen.cx = cx, m_screen.cy = cy;
}

//=========================================================================//
//                                                                         //
//     Calculates the faces of the Rubik's Cube in a continuous movement   //
//                      the whole cube or a deformation                    //
//                                                                         //
CRect CCubes::Comput(t_Mask mask)         // deformation order of the cube //
//                                        // defines the relevant tranche  //
//=========================================================================//
{  // calculate the position of the cubes during a continuous movement
   t_Mask cmp = (mask & 0x0100)? 0: (mask & 0x3F) << 9, msk = mask & 0x7E00;
   float size = float(m_screen.cx), yObs = float(*m_pYObs);
   CPoint org(m_screen.cx / 2, m_screen.cy / 2);
   CMatrix Mmo(m_pMro, m_pMmr);  // orientation matrix of the moving cubes
   CRect boxSave(m_box);
   
   // loop on the calculation of each cubie
   m_nbSide = 0, m_box.SetRectEmpty();
   for (BYTE noCube = 0; noCube < 27; noCube++) {
      t_Mask cube = m_cube[noCube];
      // plan selection cubies (fixed or mobile)
      const CMatrix *pMeo = (mask && (cube & msk) == cmp)? &Mmo: m_pMro;

	   // calculation of the 8 vertices of the cubie in 2D perspective
      CCubeVertexList vrtxList(cube, pMeo, m_dim, size, yObs);

	   // calculation of the 6 faces of the cubie and memorizing visible
      for (BYTE noSide = 0; noSide < 6; noSide++)
         if (m_nbSide < 81 // au maximum, 3 faces par cube sont visibles
          && m_side[m_nbSide].Create(vrtxList, noSide, cube, noCube, org)) {
            // the face is visible ie facing the viewer 
		      m_box |= m_side[m_nbSide++].m_box;
         }
   }

   // classification of visible faces in z (Z buffer) for masking relative
   SZSort(0, m_nbSide - 1);

   return boxSave | m_box;                // rectangle de modification dessin
}

//=========================================================================//
//                                                                         //
//          Calculates the state of Rubik's Cube after a deformation       //
//                   and resets the deformation matrix                     //
//                                                                         //
CCubes CCubes::Fixe(t_Mask mask,      // deformation order                   //
                  bool sort,        // enable sort cubies                  //
                  t_Matrix mx)      // new orientation of mobile part      //                                               //
//=========================================================================//
{           // fixes the position of the cubes was about 90 degrees
   t_Mask cmp = (mask & 0x0100)? 0: (mask & 0x3F) << 9, msk = mask & 0x7E00;
   if (mx == 0) mx = GetQMmr();            // deformation matrix

   for (BYTE noCube = 0; noCube < 27; noCube++) {
      t_Mask cube = m_cube[noCube];       // reading the old state
      if ((cube & msk) != cmp) continue;  // checks if update

      t_Matrix co = MatrixProduct(mx, (t_Matrix)cube);
      t_Mask pos = VectorOrientation(cube, mx);
      m_cube[noCube] = pos | co;
   }

   if (sort) CubeSort(0, 26);             // cubies sorting

   if (m_pMmr) *m_pMmr = CMatrix();       // reinitit the matrix
   m_oldMove = mx;                        // save matrix value for the debug

   return *this;
}

//=========================================================================//
//                                                                         //
// Calculate the faces of the developed representation of the Rubik's Cube //
//                                                                         //
CRect CCubes::Comput2D(int mod,           // cube sizing module            //
                       CPoint ofs,        // offset for the display        //
                       float ds,          // demi cote d'une face en 2D    //
                    const t_Side dscr[6]) // display sides table           //
//=========================================================================//
{
   CRect rect(ofs.x, ofs.y, ofs.x, ofs.y);
   BYTE noSide, noCube, outSide;
   float fmod = float(mod);

   // extract the faces corresponding to each page descriptor, they
   // correspond to the faces of the Rubik's Cube
   for (noSide = 0, outSide = 0; noSide < 6; noSide++) {
      const t_Side &side = dscr[noSide];
      if (side.plan == 0) continue;       // empty page
      CPoint org((side.x * 3 * mod)/ 2 + ofs.x, (side.y * 3 * mod)/ 2 + ofs.y);
      CMatrix mos(t_Matrix(side.plan));   // orientation face matrix
      
      for (noCube = 0; noCube < 27; noCube++) {
         t_Mask cube = m_cube[noCube];

         // isolate the cubies belonging to the face defined by the descriptor
         if (CubeChk(cube, side.plan)) {
            // the cubic corresondant the descriptor is found,
            // (9 cubies must correspond to this definition)
            ASSERT(outSide < 54);         // 6 plans x 9 cubies

            // calculate the corresponding surface
            if (m_si2D[outSide].Create2D(mos, cube, noSide, org, fmod, ds))
               rect |= m_si2D[outSide++].m_box;
         }
      }
   }

   return rect;
}

//=========================================================================//
//                                                                         //
// Calculate the faces of the developed representation of the Rubik's Cube //
//                                                                         //
CRect CCubes::Comput2D(int mod,           // cube sizing module            //
                       CPoint ofs,        // position of display           //
                       float ds)          // demi cote d'une face en 2D    //
//=========================================================================//
{
   return Comput2D(mod, ofs, ds, dscr);
}

//=========================================================================//
//                                                                         //
//            Drawing of the Rubik's Cube (calculate by Comput)            //
//                                                                         //
void CCubes::Draw(CDC *pDC)                                                //
//=========================================================================//
{
   CBrush *pBrush = pDC->GetCurrentBrush();

   for (BYTE i = 0; i < m_nbSide; i++)
      m_side[i].m_pSide->Draw(pDC);

   pDC->SelectObject(pBrush);
}

//=========================================================================//
//                                                                         //
//          Drawing the Rubik's Cube faces (calculate by Comput2D)         //
//                                                                         //
void CCubes::Draw2D(CDC *pDC)                                              //
//=========================================================================//
{
   CBrush *pBrush = pDC->GetCurrentBrush();

   pDC->SetBkMode(TRANSPARENT);
   for (CSide *pS = m_si2D; pS < m_si2D + 54 && pS->m_pSide; pS++)
      pS->Draw(pDC), pS->DrawId(pDC);

   pDC->SelectObject(pBrush);
}

//=========================================================================//
//                                                                         //
//             Calculate and draw a face of the Rubik's Cube               //
//                                                                         //
void CCubes::DrawSide(CDC *pDC,           // CDC to ouput                  //
                      BYTE noSide,        // no de la face                 //
                      int size,           // taille de la face             //
                      const CPoint &org)  // offset image                  //
//=========================================================================//
{
   CBrush *pBrush = pDC->GetCurrentBrush();
   CCubes cubes = *this;
   t_Side dscrIn[6];

   // Calculate the facets of the face defined
   memset(dscrIn, 0, sizeof(dscrIn));
   dscrIn[0].plan = dscr[noSide].plan;
   dscrIn[0].x = dscrIn[0].y = 1;
   cubes.Comput2D(size / 3, org, 0.45f, dscrIn);

   // draw the 9 facets composing the face
   for (UINT i = 0; i < 9 && cubes.m_si2D[i].m_pSide; i++)
      cubes.m_si2D[i].Draw(pDC);

   pDC->SelectObject(pBrush);
}

//=========================================================================//
//                                                                         //
CVertex CCubes::CenterOfSlide(BYTE noSlide) const                          //
//=========================================================================//
{
   return CVertex();
}

//=========================================================================//
//                                                                         //
//  Determining the calculated face by 'Comput', designed by the position  //
//                                                                         //
BYTE CCubes::GetSide(const CPoint &pos) const                              //
//=========================================================================//
{
   int i;

   for (i = m_nbSide; --i >= 0; )
      if (m_side[i].m_pSide->PtInSide(pos)) break;

   return BYTE(i);
}

//=========================================================================//
//                                                                         //
//              Determining the cubie designed by the position             //
//                                                                         //
BYTE CCubes::GetCubeNo(const CPoint &pos) const                            //
//=========================================================================//
{
   for (int i = m_nbSide; --i >= 0; )
      if (m_side[i].m_pSide->PtInSide(pos))
         return m_side[i].GetCubeNo();

   return 0xFF;
}

//=========================================================================//
//                                                                         //
// Determining the calculated face by 'Comput2D', designed by the position //
//                                                                         //
BYTE CCubes::GetSide2D(const CPoint &pos) const                            //
//=========================================================================//
{
   for (BYTE i = 0; i < 54; i++)
      if (m_si2D[i].m_pSide->PtInSide(pos)) return i;

   return 0xFF;
}

//=========================================================================//
//                                                                         //
//         Calculate the deformation applied to the Rubik's Cube           //
//                                                                         //
t_Command CCubes::GetMove(int x, int y,   // Moving the mouse              //
                          BYTE cubeNo)    // selected cubie                //
//=========================================================================//
{// calculate the change order of the cube (according to a mouse movement)
   float dx, dy, dz;
   BYTE axe;

	// be put in the coordinate of the object
   dx = m_pMro->m_d[0][0] * y + m_pMro->m_d[0][2] * x;
   dy = m_pMro->m_d[1][0] * y + m_pMro->m_d[1][2] * x;
   dz = m_pMro->m_d[2][0] * y + m_pMro->m_d[2][2] * x;

	// determine the axis of motion to execute
   short nx = short(dx), ny = short(dy), nz = short(dz), n;
   if (dx < 0.0f) dx = -dx;
   if (dy < 0.0f) dy = -dy;
   if (dz < 0.0f) dz = -dz;
                         n = nx, axe = 0; // axe =>x
   if (dx < dy) dx = dy, n = ny, axe = 1; // axe =>y
   if (dx < dz) dx = dz, n = nz, axe = 2; // axe =>z

   t_Command cmd = CCmdString::toCommand(axe, cubeNo);
   if (cmd == 0) return cmd;
   t_Command o = (cmd & 0x1F)/ 6;
   if (n < 0) o = 4 - o;

   return (cmd & 0x60) + (cmd & 0x1F)% 6 + o * 6;
}

//=========================================================================//
//                                                                         //
//               Rank facets according to their distance z                 //
//                to draw the most distant and the first                   //
//                   and deal the problems of masking                      //
//                                                                         //
void CCubes::SZSort(int low, int hig)                                      //
//=========================================================================//
{
   ASSERT(low >= 0 && hig < m_nbSide && low < hig);
   // extract the z of the middle element
   int mz = m_side[(low + hig) / 2].m_pSide->m_z;
   int lo = low, hi = hig;

   // loop until the intersection index
   while ( lo <= hi ) {
      // look at the top and at the bottom elements not in place
      while ( lo < hig && mz > m_side[lo].m_pSide->m_z ) ++lo;
      while ( hi > low && mz < m_side[hi].m_pSide->m_z ) --hi;

      // exchange elements encountered
      if ( lo <= hi ) m_side[lo++].Swap(m_side + hi--);
   }

   // recursive achieve the sort of remains
   if ( low < hi  ) SZSort( low, hi );
   if ( lo  < hig ) SZSort( lo, hig );
}

//=========================================================================//
//                                                                         //
//               Classification, depending on their position,              //
//                      faces calculated by Comput2D                       //
//                                                                         //
void CCubes::SPSort(int low, int hig)                                      //
//=========================================================================//
{
   ASSERT(low >= 0 && hig < 54 && low < hig);
   // Extrait le z de l'element median
   CSide *pMs = m_si2D[(low + hig) / 2].m_pSide;
   int lo = low, hi = hig;

   // loop until the intersection index
   while ( lo <= hi ) {
      // look at the top and at the bottom elements not in place
      while ( lo < hig && pMs->Compare(m_si2D[lo].m_pSide) > 0 ) ++lo;
      while ( hi > low && pMs->Compare(m_si2D[hi].m_pSide) < 0 ) --hi;

      // echange les elements rencontres
      if ( lo <= hi ) m_si2D[lo++].Swap(m_si2D + hi--);
   }

   // recursive achieve the sort of remains
   if ( low < hi  ) SPSort( low, hi );
   if ( lo  < hig ) SPSort( lo, hig );
}

//=========================================================================//
inline static void CubeSwap(int i, int j, t_Mask cube[])                   //
   { t_Mask t = cube[i]; if (i != j) cube[i] = cube[j], cube[j] = t; }     //
//=========================================================================//
//                                                                         //
//                            Cubies sorting                               //
//                                                                         //
void CCubes::CubeSort(int low, int hig)                                    //
//=========================================================================//
{
   t_Mask m = m_cube[(low + hig) / 2];
   int lo = low, hi = hig;

   // loop until the intersection index
   while ( lo <= hi ) {
      // look at the top and at the bottom elements not in place
      while ( lo < hig && m > m_cube[lo] ) ++lo;
      while ( hi > low && m < m_cube[hi] ) --hi;

      // echange les elements rencontres
      if ( lo <= hi ) CubeSwap( lo++, hi--, m_cube );
   }

   // recursive achieve the sort of remains
   if ( low < hi  ) CubeSort( low, hi );
   if ( lo  < hig ) CubeSort( lo, hig );
}

//=========================================================================//
//                                                                         //
//       Conversion of the state of Rubik's Cube in an ASCII string        //
//                                                                         //
CString CCubes::GetText(                                                   //                                   
                t_Matrix or,           // New Direction (optional)         //
                const CString &colors  // 6 color letters + 6 face letters //
                        ) const        // or + 1 or 3 to facets rotations  //
//=========================================================================//
{
   TCHAR color;
   CString text;
   DWORD centers;
   CCubes cubes(this);
   int len = colors.GetLength();

   ASSERT(len == 12 || len == 13 || len == 15);

   // calculate the new direction
   if (or > 1) cubes.Fixe(0, true, or);   // set the new direction

   // Construction and ordering of the held flat
   cubes.Comput2D(50, CPoint(0,0), 0.45f, dscrText), cubes.SPSort();

   // calculates the center of side rotations
   centers = (len > 12)? cubes.CenterSidesRotation(): 0;

   // for each face of the Rubik's Cube
   for (int i = 0, k = 0; i < 6; i++) {
      color = colors.GetAt(cubes.m_si2D[k].m_pSide->m_noDscr + 6);
      text += color;
      text += TCHAR(':');

      // color output of the 9 facets
      for (int j = 0; j < 9; j++, k++)
         text += colors.GetAt( cubes.m_si2D[k].m_pSide->m_noSide );

      // add the rotation of center of face id (or space)
      if (len == 15)
         CenterSidesRotationAdd(text, centers, color, colors.Right(3));
      else
         CenterSidesRotationAdd(text, centers, color);
   }
   text.TrimRight();
   
   return text;
}

//=========================================================================//
//                                                                         //
//            Calculate the 3 colors of the centers of the faces           //
//                                                                         //
t_GROUP CCubes::SidesColors(bool ref) const  // reference or this Cube     //
//=========================================================================//
{
   if (ref) {

      return RefColors(CENTER_FACES);

   } else {
      BYTE facets[54];

      SetSideText((*this).GetText(), facets, "ORGBWYUDLRFB");

      return GetSideColors(CENTER_FACES, facets);
   }
}

//=========================================================================//
//                                                                         //
//          Initialize a Rubik's Cube from a table of 54 colors            //
//                                                                         //
int CCubes::Create(BYTE facets[54], // table of 54 facets colors           //
                   bool er)         // true : erase the facets in error    //
//=========================================================================//
{
   CCubes cubes;
   BYTE noCube, i;
   UINT mask = 0;

   // calculate the orientation of the Rubik's Cube
   t_Matrix or = GetCubeOrientation(facets);
   if (or == 0)
      return -8;    // central facets in error, incorrectly positioned

   for (i = 0; i < 27; i++) {
      // facets colors of the Cube
      DWORD nSides = getNoSides(i);
      DWORD colors = GetSideColors(nSides, facets);
      // Cube orientation
      t_Matrix cOr = GetCubeOrientation(RefColors(nSides), colors, or);

      if ( cOr && (noCube = GetNoCube(cvrtColors(colors))) < 27
        && (mask & (1 << noCube))== 0 ) {
         // to test the presence of all cubies and only one
         mask |= 1 << noCube;

         // coding of the cubie
         cubes.m_cube[noCube] = cOr|cubePos(i);
      } else if (er) {
         SetSideColors(nSides, GROUP_EMPTY, facets);
      }
   }

   if (mask != (1 << 27)- 1)
      return -9;     // Error, cubies are undefine

   cubes.SetCenterSidesRotation(facets);

   // ok, Rubik's Cube update
   *this = cubes;
   return 0;
}

//=========================================================================//
//                                                                         //
//                    Read a Rubik's Cube from a string                    //
//                                                                         //
BOOL CCubes::StringToCube(const CString &text)                             //
//=========================================================================//
{
   int type = FileTypeSelect(text), er = 0;

   if (type == TFMT_RBX) {
      // internal format RubixCube
      CCubes cube;

      for (int i = 0; i < 27; i++)
         cube.m_cube[i] = cubePos(i) | text[i + 6] - TCHAR('A') + 1;

      if (TCHAR(cube.GetHash() % 26 + 'A') == text[5]) {
         *this = cube;
         return TRUE;
      }
   } else if (type == TFMT_COLORS) {
      // internal format colors

      if (ReadColors(text))
         return TRUE;

   } else if (type == TFMT_POS) {
      // format "L:OWGGORWGB F:YROGGOWRR R:BGRYRBBYG B:WRGOBWOWR...
      BYTE facets[54];

      if ( (er = SetSideText(text, facets, "ORGBWYUDLRFBo"))== 0
        && (er = Create(facets, false))== 0 )
         return TRUE;

   } else if (type == TFMT_SGM) {
      // format Singmaster
      BYTE facets[54];

      if ( SingmasterToFacets(text, facets)
        && (er = Create(facets, false))== 0 )
         return TRUE;

   } else if (type == TFMT_MAN) {
      // format maneuvre

      // preserves the initial orientation of the Rubik's Cube, and turn
      // the faces in relation to this orientation
      Init(Orientation());
      CmdApply( CCmdString().CmdExtract(text) );
      return TRUE;
   }
   return FALSE;
}

//=========================================================================//
//                                                                         //
//                    Read a Rubik's Cube from a file                      //
//                                                                         //
BOOL CCubes::Read(const CString &fileName, CString *pComments)             //
//=========================================================================//
{
   static const CString test("/(\n\r");
   CString text = ReadText(fileName);
   int i, ix;

   if (pComments) {
      // extract end of command string to copy in comments
      for (i = 0; i < text.GetLength() - 1; i++) {
         if ((ix = test.Find(text.GetAt(i))) >= 0) {
            if (ix < 2)
               *pComments = text.Mid(i);
            else if (text.GetAt(i + 1) == '\r' || text.GetAt(i + 1) == '\n')
               *pComments = text.Mid(i + 2);
            else 
               *pComments = text.Mid(i + 1);
            break;
         }
      }
   }

   return StringToCube(text);
}

//=========================================================================//
//                                                                         //
//         Analyse du contenu d'un fichier pour determiner le type         //
//                                                                         //
int CCubes::FileTypeSelect(const CString &msg)                             //
//=========================================================================//
{
   // test format RubixCube
   if (msg.Left(5) == TEXT("RUBIK") && msg.GetLength() > 32)
      return TFMT_RBX;

   // test format Color
   if (msg.Left(6) == TEXT("COLORS") && msg.GetLength() > 32)
      return TFMT_COLORS;

   static const CString test(":'URFDLB(/[{;\\\n\r");
   DWORD counts = 0;
   int i, ix;

   // calculate the signature of the content (first line)
   for (i = 0; i < msg.GetLength(); i++)
      if ( (ix = test.Find(msg.GetAt(i)))>= 8 ) break; // end of data
      else if (ix >= 0) counts += 1 << (ix * 4);

   //   TRACE("CCubes::FileTypeSelect %08x\n", counts);
   if (counts == 0xA111A106 || counts == 0x11111106)
      return TFMT_POS;        // "L:OWGGORWGB F:YROGGOWRR R:BGRYRBBYG ... 
   else if (counts == 0x88888800)
      return TFMT_SGM;        // Singmaster format
   else if ((counts & 0x0F) == 0 && counts != 0)
      return TFMT_MAN;        // maneuver notation
   else
      return TFMT_NULL;
}

//=========================================================================//
//                                                                         //
//          Analyse du nom d'un fichier pour determiner le type            //
//                                                                         //
int CCubes::FileExtSelect(const CString &fileName)                         //
//=========================================================================//
{
   CString ext = fileName.Right(4);

   ext.MakeLower();

   if (ext == ".rbx") return TFMT_RBX;
   if (ext == ".col") return TFMT_COLORS;
   if (ext == ".txt"
    || ext == ".pos") return TFMT_POS;
   if (ext == ".sgm") return TFMT_SGM;
   if (ext == ".man"
    || ext == ".cmd") return TFMT_MAN;
   if (ext == ".bmp" || ext == ".jpg" || ext == ".gif"
    || ext == ".png") return TFMT_BMP;
   if (ext == ".tif") return TFMT_TIF;

   return TFMT_NULL;
}

//=========================================================================//
//                                                                         //
//               Convert a Rubik's Cube state in a string                  //
//                                                                         //
CString CCubes::ConvertString(int type)                                    //
//=========================================================================//
{
   if (type == TFMT_POS) {
      // format position
      return GetText();

   } else if (type == TFMT_SGM) {
      // format singmaster
      static const CString colorsId("ORGBWYUDLRFB");
      BYTE facets[54];

      SetSideText(GetText(), facets, colorsId);
      return ToSingmasterString(facets, colorsId);

   } else if (type == TFMT_COLORS) {
      return WriteColors();

   } else {
      // write internal format RubixCube
      CCubes cube(this);
      CString text;

      // header with check sum
      cube.CubeSort(0, 26);
      text.Format("RUBIK%c", cube.GetHash() % 26 + 'A');

      // one letter for each cubie orientation ('A' to 'X' -> 1 to 24)
      for (int i = 0; i < 27; i++) text += TCHAR(cube[i] + 'A' - 1);
      return text;
   }
}

//=========================================================================//
//                                                                         //
//                   Writing a Rubik's Cube in a file                      //
//                                                                         //
BOOL CCubes::Write(const CString &fileName, const CString &comments)       //
//=========================================================================//
{
   CString text = ConvertString(FileExtSelect(fileName)) + "\r\n";
   CFile file;

   text += comments;
   if (file.Open(fileName, CFile::modeCreate|CFile::modeWrite))
      file.Write(text.GetBuffer(0), text.GetLength());

   return TRUE;
}

//=========================================================================//
//                                                                         //
//                   Execute the command of deformation                    //
//                                                                         //
CCubes CCubes::CmdApply(const CCmdString &cmd)   // command string         //
//=========================================================================//
{
   int length = cmd.GetLength();

//   TRACE("CCubes::CmdApply %s (%d)\n", cmd.CmdConvert(), length);

   for (int i = 0; i < length; i++) {
      t_Command ci = cmd.GetAt(i);

      if (ci == 1) {
         continue;
      } else if (ci == 4) {
         ASSERT(0);
      } else if (ci == 5) {
         // initialization with a series of values

         ASSERT(length - i >= 27);

         for (int j = 0; j < 27; j++)
            m_cube[i] = cubePos(i) | cmd.GetAt(i + j + 1);

         i += 27;
      } else if (ci >= 6) {
         t_Mask move = CmdCubeSet(ci);
         t_Matrix ma = CmdToRotate(ci);
//         mob.CmdSet(ci);

         Fixe(move, false, ma);
      }
   }

   CubeSort(0, 26); // cubies sorting

   return *this;
}

//=========================================================================//
//                                                                         //
//                    Storing the history of commands                      //
//                                                                         //
void CCubes::CmdAddRevers(t_Command cmd,  // new commande                  //
                      CCmdString &cmdLine,// history                       //
                      DWORD curs)         // source position of command    //
//=========================================================================//
{  
   if (cmd == 1) {
      // skip a line

      cmdLine.Insert(0, cmd);
   } else if (cmd == 4) {
      // reset the Rubik's Cube
      CCubes cubes(*this);
      int i;

      cubes.CubeSort(0, 26);
      for (i = 0; i < 27; i++) cmdLine.Insert(i, BYTE(cubes.m_cube[i]) + 0x60);
      cmdLine.Insert(0, 5);
   } else if (cmd == 5) {
      ASSERT(0);
   } else if (cmd & 0x80) {
      // orientation of the Rubik's Cube

      cmdLine.Insert(0, cmd ^ 0x40);
   } else if (cmd >= 6) {
      // deformation of the Rubik's Cube
      TCHAR flag = cmd & 0xE0, c = cmd & 0x1F;

      cmdLine.Insert(0, TCHAR((4 - c / 6)* 6 + c % 6 + flag));
   }

   if (cmd && curs != DWORD(-1)) {
      // save the source position of command
      CString tmp;
      tmp.Format("\2%08X", curs);
      cmdLine.Insert(0, tmp);
   }
}

//=========================================================================//
//                                                                         //
//                    Extracting the commands of history                   //
//                                                                         //
t_Command CCubes::CmdExtract(CCmdString &cmdLine,  // history              //
                             DWORD *pCurs)         // source id of cmd     //
//=========================================================================//
{
   int len;

   while ((len = cmdLine.GetLength()) > 0) {
      t_Command cmd = cmdLine.GetAt(0);

      cmdLine.Delete(0, 1);

      if (cmd == 2) {
         if (len >= 8) {
            if (pCurs) *pCurs = strtoul(cmdLine.Left(8), NULL, 16);
            cmdLine.Delete(0, 8);
         } else {
            cmdLine.Empty();
         }
      } else if (cmd == 5) {
         if (len >= 27) {
            for (int i = 0; i < 27; i++)
               m_cube[i] = cubePos(i) | (cmdLine.GetAt(i) - 0x60);

            cmdLine.Delete(0, 27);
         } else {
            cmdLine.Empty();
         }
         return 0;
      } else if (cmd >= 6) {
         return cmd;
      }
   }
   return 0;
}

//=========================================================================//
//                                                                         //
//       Test whether the colors of edges and vertices are compatible      //
//               and also that the Rubik's Cube can be ordered             //
//                                                                         //
int CCubes::ColorsCheck() const                                            //
//=========================================================================//
{
   BYTE no, np, center[1], centerFace[6], edge[12], corner[8], facets[54];
   BYTE cp[4] = {0,0,0,0}, *pTb[4] = {center,centerFace,edge,corner};
   t_Matrix or = MatrixInvert(Orientation());
   t_CUBIES pres = 0;

   // fill the facet table, the cube moved into position origin for
   // calculating the parity of permutations of places of the cubies
   int er = SetSideText(GetText(or), facets, "ORGBWYUDLRFB");
   if (er < 0) {
      return er;
   }
   if (GetCubeOrientation(facets)!= 1) {
      return -8;        // error : incorrectly positioned central facets
   }
   // examine the colors of the facets of the 27 cubies
   for (no = 0; no < 27; no++) {
      t_GROUP nSides = getNoSides(no);
      t_GROUP colors = GetSideColors(nSides, facets);
      BYTE nbc = SidesCount(nSides);
      BYTE noCube;      // cube corresponding to the color signature
      char orient = CheckSidesColors(colors, &noCube);

      if (orient < 0) {
         return orient;    // error : color illegal or disordered
      }
      // count the number of rotations of the color
      cp[nbc] += orient;

      // mark the presence of cubies
      pres |= cvrtCubie(noCube);
      *pTb[nbc]++ = noCube;
   }

   if (pres != ALL_CUBIES) {
      return -9;        // error : cubes are duplicated, others lack
   }
   // A negative error indicates that it is materially impossible to build
   // a Rubik's Cube with the colors of facets. It should only occur in a
   // program to create, not in a test.

   // checking program and content of facets tables
   ASSERT(pTb[0] == center + 1 && pTb[1] == centerFace + 6
             && pTb[2] == edge + 12 && pTb[3] == corner + 8);

   // calculate the parity of the permutations of locations of the cubies 
   np = PermutationsParity(edge, 12) ^ PermutationsParity(corner, 8);

   // Possible errors of consistency of orientation and position of the
   // cubies
   return (cp[2] % 2? 1: 0) + (cp[3] % 3? 2: 0) + (np? 4: 0);
   // A positive error indicates that it is not possible to order the
   // Rubik's Cube
}

//=========================================================================//
//                                                                         //
//                    Calculate the state of Rubik's Cube                  //
//                                                                         //
DWORD CCubes::GetHash() const                                              //
//=========================================================================//
{
   CCubes cubes(this);
   DWORD h = 0, i;

   cubes.CubeSort(0, 26);
   for (i = 0; i < 21; i++) h += h * 2 + cubes.m_cube[ moveCubes[i] ];

   return h;
}

//=========================================================================//
//                                                                         //
//                    Determine the cubic misaligned                       //
//                                                                         //
t_CUBIES CCubes::ScrambledCubies() const                                   //
//=========================================================================//
{
   t_Matrix or = Orientation();
   t_CUBIES cubeBits = 0;      // bits non-aligned cubes
   CCubes cube(this);

   cube.CubeSort(0, 26);
   // search cubies no correctly oriented
   for (int i = 0; i < 27; i++)
      if ( t_Matrix(cube[i]) != or ) cubeBits |= cvrtCubie(i);

   return cubeBits;
}

//=========================================================================//
//                                                                         //
//                   Calculate the state of Rubik's Cube                   //
//                                                                         //
BYTE CCubes::GetState() const                                              //
//=========================================================================//
{
   static const t_CUBIES cf = GetCubiesVSC(1);// bits cubes centers of faces
   t_CUBIES cubeBits = ScrambledCubies();     // set bits non-aligned cubes

   // 0x02 => center faces scrambled, 0x01 => Rubik's Cube scrambled
   return ((cubeBits & cf)? 2: 0) | ((cubeBits & ~cf)? 1: 0);
}


// cube with one facet     R  L  B  F  U  D
static const BYTE cf[6] = {14,12,16,10,22,4}; // cubies numbers
static const CString codeCf("RLBFUD");
//=========================================================================//
//                                                                         //
//         Calculate the rotation state of the facets center side          //
//                      of the Rubik's Cube                                //
//                                                                         //
DWORD CCubes::CenterSidesRotation() const                                  //
//=========================================================================//
{
   t_Matrix or = MatrixInvert(Orientation()), co;
   BYTE cube, face, q;
   CCubes cubes(this);
   DWORD centers = 0;

   cubes.Fixe(); cubes.CubeSort(0, 26);

   // for the six sides
   for (face = 0; face < 6; face++) {
      cube = cf[face];                    // cubie in center of side
      co = MatrixProduct((t_Matrix)cubes[cube], or);
      q = toQarterTurn(co) >> (face & ~1);// shift 0, 2, 4
      if ((q &= 3) > 0) {                 // cubie orientation in plan of side
         if (face & 1) q = 4 - q;         // negate on 1, 3, 5
         centers |= q << (face * 4);
      }
   }
      
   return centers;
}

//=========================================================================//
//                                                                         //
//            Decoding facets rotations of the center sides                //
//                                                                         //
void CCubes::SetCenterSidesRotation(const BYTE facets[54])                 //
//=========================================================================//
{
   t_Matrix fo, or = Orientation();
   BYTE cube, face, q;

   CubeSort(0, 26);
   for (face = 0; face < 6; ++face) {
      cube = cf[face];

      // facet rotation
      q = facets[ BYTE(getNoSides(cube)) ] >> 4;
      if (q > 0 && q < 4) {
         // update the rotation, depending the facet number
         if (face & 1) q = 4 - q;               // negate on 1, 3, 5
         fo = MatrixOrient(q << (face & 0xE));  // 00zzyyxx to t_Matrix

         // update the Rubik's Cube
         m_cube[cube] = (m_cube[cube] & 0xFF00) | MatrixProduct(fo, or);
      }
   }
}

//=========================================================================//
//                                                                         //
//      Calculate the rotation state message of the facets center side     //
//          decode the center face rotations in a position message         //
//                                                                         //
DWORD CCubes::CenterSidesRotation(const CString &msg) const                //
//=========================================================================//
{
   int i, ix, len = msg.GetLength() - 11;
   TCHAR idFace[] = " :";
   DWORD centers = 0;

   for (i = 0; i < 6; i++) {
      idFace[0] = codeCf[i];
      ix = msg.Find(idFace, 0);
      if (ix >= 0 && ix < len) {
         BYTE or = 0;
         switch (msg[ix + 11]) {
         case '1' : case '>' : case '+' : or = 1; break;
         case '2' : case '*' : case '#' : or = 2; break;
         case '3' : case '<' : case '-' : or = 3; break;
         }
         centers |= or << (i * 4);
      }
   }
   return centers;
}

//=========================================================================//
//                                                                         //
//      Calculate the rotation state message of the facets center side     //
//          code the center face rotation in position message              //
//                                                                         //
void CCubes::CenterSidesRotationAdd(CString &text, // text to modify       //
                                    DWORD centers, // rotation of centers  //
                                    TCHAR face,    // face letter          //
                             const CString &model) // 3 char to add          //
                                          const                            //
//=========================================================================//
{
   CString rotChar = CString(' ') + model;
   int ix = codeCf.Find(face, 0);

   if (ix >= 0)
      text += rotChar.GetAt( (centers >> (ix * 4)) & 3 );
}

//=========================================================================//
//                                                                         //
//      Calculate the rotation state message of the facets center side     //
//                      of the Rubik's Cube                                //
//                                                                         //
CString CCubes::GetCentersMsg(const CString &order) const// facets rotation//
//=========================================================================//
{
   DWORD centers = CenterSidesRotation();
   CString msg = "C:";              // => Centers

   for (int i = 0, ix; i < 6; i++) {
      ix = codeCf.Find(order.GetAt(i));
      ASSERT(ix >= 0 && ix < 6);
      msg += TCHAR( '0' + ((centers >> (ix * 4)) & 3) );
   }

   return msg;
}

//=========================================================================//
//                                                                         //
//                          Debug : Dump parameters                        //
//                                                                         //
CString CCubes::Dump(int mode) const                                       //
//=========================================================================//
{
   CString msg, tmp;
   int i, j, k;

   switch(mode & 0x000F) {
   case 0 :
      msg = "CCubes::Dim\n";
      tmp.Format(" dim=%.4f z=%d\n", double(m_dim), *m_pYObs);
      msg += tmp;
      break;
   case 1 :
      msg = "CCubes::cubes\n";
      for (i = 0; i < 3; i++) {
         for(j = 0; j < 3; j++) {
            k = j*3 + i*9;
            tmp.Format(" %04X-%04X-%04X", m_cube[k], m_cube[k + 1], m_cube[k + 2]);
            msg += tmp;
         }
         msg += "\n";
      }
      msg += CMatrix::DumpMat(Orientation(), "Center") + "\n";
      tmp.Format("status = %d orient : %s",
                     GetState(), GetMsgCenter(CenterSidesRotation(), -1));
      msg += tmp;
      break;
   case 2 :
      msg = CString("Matrix Mro (orientation)\n") + m_pMro->Dump(3);
      break;
   case 3 :
      msg = CString("Matrix Mmr (deformation)\n") + m_pMmr->Dump(3);
      msg += CMatrix::DumpMat(m_oldMove, " Mmr old") + "\n";
      break;
   case 4 :
      msg = "CCubes::graph\n";
      tmp.Format(" box(%d,%d) side=%d\n", m_box.Width(), m_box.Height(), m_nbSide);
      msg += tmp;
      break;
   case 5 :
      msg.Format("CCubes facets colors\n%s\nHash code = %08X\n", GetText(), GetHash());
      break;
   case 6 :
      msg = "$6"; // pour forcer l'affichage de la representation developpee
      break;
   case 7 :
      for (i = 0; i < 6; i++) {
         t_Matrix m = t_Matrix( dscr[i].plan );
         tmp.Format(" Face_%d %02X :", i, m);
         msg += CMatrix::DumpMat( m, tmp );
         msg += "\n";
      }
      break;
      // pour terminer l'affichage
   default: msg = "$$"; // hide msg
   }
   return msg;
}
/////////////////////////////////////////////////////////////////////////////
static t_GROUP sideColorsTable[25] = { DWORD(-1), DWORD(-1), };
static t_Matrix cubeOr[6][6];

void InitColorsRotateTable()
{
   for (int i = 0; i < 6; i++)
      for (int j = 0; j < 6; j++) cubeOr[i][j] = 0;

   for (t_Matrix or = 1; or < 25; or++) {
      CCubes cubes;

      // Cube with the new orientation
      cubes.Fixe(0, true, or);
      
      t_GROUP sidesColors = cubes.SidesColors(false);
      sideColorsTable[or] = sidesColors;

      BYTE Upper = GetCenterColor(sidesColors, 4);  // U
      BYTE Front = GetCenterColor(sidesColors, 3);  // F

      ASSERT(cubeOr[Upper][Front] == 0);
      // update the 'Upper-Front' table
      cubeOr[Upper][Front] = or;

//      TRACE("InitColorsRotateTable up=%d fr=%d or=%d\n", Upper, Front, or);
   }
}

/////////////////////////////////////////////////////////////////////////////

BYTE GetCubeSideColor(t_Matrix or, BYTE face)
{
   ASSERT(BYTE(or - 1) < 24);
   return GetCenterColor(sideColorsTable[or], face);
}

/////////////////////////////////////////////////////////////////////////////
// By convention, the orientation of a cube is defined by the color of 2
// sides, the front and top face. This program allows rotation of a cube
// placed in the desired direction. The original cube, defined by the unitary
// matrix (1,0,0) (0,1,0) (0,0,1), presented before the face 3 (default blue
// B) and top face 4 (default W White). Position 34 (default BW) is the
// original position.

t_Matrix GetCubeOrientation(BYTE Upper, BYTE Front)
{
   return cubeOr[Upper][Front];
}

//=========================================================================//
// The variable "sides"of type DWORD contains in both endian color both    //
// sides reference orientation. It will put the cube in the original       //
// position if desired or "neg" to bring a cube oriented to the original   //
// position.                                                               //
//                                                                         //
t_Matrix GetCubeOrientation(t_GROUP sides, bool neg)                       //
//                                                                         //
//=========================================================================//
{
   BYTE Front = BYTE_0(sides), Upper = BYTE_1(sides);
   t_Matrix or = cubeOr[Upper][Front];

   return neg? MatrixInvert(or): or;
}

//=========================================================================//
//                                                                         //
//    Determine the orientation of a cubie from the colors of two facets   //
//               and reference colors of these two facets                  //
//                                                                         //
t_Matrix GetCubeOrientation(t_GROUP colors,        // facets colors        //
                               t_GROUP noSide,     // reference colors     //
                                     t_Matrix def) // default value        //
//=========================================================================//
{
   if (sideColorsTable[1] == DWORD(-1)) InitColorsRotateTable();
   if (BYTE(noSide >> 8) < 6) {
      // only two or three facet colors

      // edge cubes or corner cubes
      t_Matrix or = MatrixProduct(GetCubeOrientation(colors, true),
                                          GetCubeOrientation(noSide, false));
      if (or == 0) return 0;
      // additional check for the corner cubies (3 facets visible)
      if (BYTE(noSide >> 16) < 6) {

         // A given cubie, at a given place, has only three possible
         // orientations of the 6 combinations of three colors. The visible
         // facets remain the same whatever the position and orientation.
         if (or != MatrixProduct(GetCubeOrientation(colors >> 8, true),
                                    GetCubeOrientation(noSide >> 8, false)))
            return 0;
      }
      return or;
   }
   return def;
}

//=========================================================================//
//                                                                         //
//              Determine the orientation of a Rubik's Cube,               //
//               thanks to the color table (central facets)                //
//                                                                         //
t_Matrix GetCubeOrientation(const BYTE facets[54])                         //
//=========================================================================//
{
   // Two central aspects are sufficient to calculate the orientation.
   // We check that the other four facets confirm this orientation.
   t_GROUP nSides0 = GROUP(40, 4,22), refColor0 = RefColors(nSides0);
   t_GROUP nSides1 = GROUP(49,13,31), refColor1 = RefColors(nSides1);
   t_GROUP color0 = GetSideColors(nSides0, facets);
   t_GROUP color1 = GetSideColors(nSides1, facets);
   t_Matrix or = GetCubeOrientation(refColor0, color0, 0);

   if (or != GetCubeOrientation(refColor1, color1, 0)) {
      return 0;            // Error : central facets incorrectly positioned
   }

   return or;
}

//=========================================================================//
//                                                                         //
t_Matrix NewOrient(const CString &frontUp,// command string, two faces char//
                   const CString &colors) // ref code string               //
//=========================================================================//
{
   union { DWORD d; char c[4]; } sides;
   ASSERT(colors.GetLength() == 6);

   // decode la couleur des faces directrices
   sides.d = DWORD(-1);
   for (int i = 0; i < frontUp.GetLength(); i++) {
      sides.c[i] = (char)colors.FindOneOf(frontUp.Mid(i, 1));
      if (sides.c[i] < 0 || sides.c[i] >= 6)
         return 0;            // erreur dans le codage orientation
   }
   return GetCubeOrientation(sides.d, true);
}

/////////////////////////////////////////////////////////////////////////////


