// Cubes.h: interface for the CCubes class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_CUBES_H__7752AAB1_D4AC_11D3_A5F0_AA6FFDA5D841__INCLUDED_)
#define AFX_CUBES_H__7752AAB1_D4AC_11D3_A5F0_AA6FFDA5D841__INCLUDED_

#include "CubeUtilities.h"
#include "Matrix.h"
#include "Sides.h"

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#define MODE_TOOLTIP 5

typedef enum {
   TFMT_NULL,TFMT_RBX, TFMT_COLORS, TFMT_POS, TFMT_SGM, TFMT_MAN,
   TFMT_BMP, TFMT_TIF
}; 


///////////////////////////////////////////////////////////////////////////////
class CCubes   // Figure composed of 27 cubes arranged in 3*3*3 (Rubik's Cube)
///////////////////////////////////////////////////////////////////////////////
{        
public:
   CCubes(const CCubes *pCubes, float dim = 1.f, int *yObs = NULL,
                           const CMatrix *pObs = NULL, CMatrix *pMob = NULL);
   CCubes(float dim = 1.f, int *yObs = NULL,
                           const CMatrix *pObs = NULL, CMatrix *pMob = NULL);

   virtual ~CCubes() {}

   inline void Init(t_Matrix or = MatrixUnit)
      { for (BYTE n = 0; n < 27; n++) m_cube[n] = CubeOrigine(n, or); }

   inline t_Matrix Orientation() const { return t_Matrix(m_cube[13]); }

   void  SetScreen(int dx, int dy);

   CRect Comput(t_Mask mask = 0);
   CCubes Fixe(t_Mask mask = 0, bool sort = true, t_Matrix or = 0);

   CRect Comput2D(int mod, CPoint ofs, float ds, const t_Side dscr[6]);
   CRect Comput2D(int mod = 15, CPoint ofs = CPoint(5, 5), float ds = 0.45f);
   void  SPSort() { SPSort(0, 53); }

   inline CRect GetSide2DRect() const { return CSide::GetRect(m_si2D, 54); }
   inline CRect GetSideRect() const { return CSide::GetRect(m_side, m_nbSide); }

   inline void SPUnSort()
      { for (int i = 0; i < 54; i++) m_si2D[i].Restore(); }

   void  Draw(CDC *pDC);
   void  Draw2D(CDC *pDC);
   void  DrawSide(CDC *pDC, BYTE noSide, int size, const CPoint &org = CPoint(0,0));

   BYTE  GetSide(const CPoint &pos) const;
   CVertex CenterOfSlide(BYTE noSlide) const;
   BYTE  GetCubeNo(const CPoint &pos) const;
   BYTE  GetSide2D(const CPoint &pos) const;

   inline CSide *GetSide(const BYTE no) const
      { return (no < m_nbSide)? m_side[no].m_pSide: NULL; }
   inline CSide *GetSide2D(const BYTE no) const
      { return (no < 54)? m_si2D[no].m_pSide: NULL; }
   inline int Si2DSortIndex(int i) const
      { return m_si2D[i].m_pSide - m_si2D; }

   t_Command GetMove(int x, int y, BYTE cubeNo);

   inline CCubes operator= (t_Matrix or)
      { Init(or); return *this; }
   inline CCubes operator= (const CCubes &cubes)
      { memcpy(m_cube, cubes.m_cube, sizeof(m_cube)); return *this; }
   inline CCubes operator= (const CCubes *pCubes)
      { memcpy(m_cube, pCubes->m_cube, sizeof(m_cube)); return *this; }
   inline t_Mask operator[] (int index) const
      { return (index >= 0 && index < 27)? m_cube[index]: 0; }

   DWORD GetHash() const;
   BYTE  GetState() const;

   CString GetText(t_Matrix or = 1,
                   const CString &colorsSides = "ORGBWYRLBFUD") const;
   CString ConvertString(int type);

   t_GROUP SidesColors(bool def) const;

   int Create(BYTE facets[54], bool er);

   BOOL StringToCube(const CString &msg);

   BOOL Read(const CString &fileName, CString *pComments = NULL);
   BOOL Write(const CString &fileName, const CString &comments = "");
   int TypeSelect(const CString &msg) const;

   static int FileTypeSelect(const CString &msg);
   static int FileExtSelect(const CString &fileName);

   CCubes CmdApply(const CCmdString &cmd);
   void CmdAddRevers(t_Command cmd, CCmdString &cmdLine, DWORD curs = DWORD(-1));
   t_Command CmdExtract(CCmdString &cmdLine, DWORD *pCurs = NULL);

   int ColorsCheck() const;

   t_CUBIES ScrambledCubies() const;
   DWORD CenterSidesRotation() const;
   CString GetCentersMsg(const CString &order = "ULFRBD") const;
   DWORD CenterSidesRotation(const CString &msg) const;
   void CenterSidesRotationAdd(CString &text, DWORD centers, TCHAR face,
                                 const CString &model = "+#-") const;

   inline CRect GetBox() const { return m_box; }

   CString Dump(int mode) const;

protected:
   inline t_Mask CubeOrigine(BYTE noCube, t_Matrix or = MatrixUnit) const
      { return cubePos(noCube) | or; }

   inline t_Matrix GetQMmr() const
      { return (m_pMmr)? m_pMmr->Pack(): MatrixUnit; }

   void CubeSort(int low, int hig);
   void SetCenterSidesRotation(const BYTE facets[54]);

   // La position des centres des cubes lmentaires d'un Rubik's Cube peut
   // tre reprsente par un vecteur compos de trois valeurs prsentant les
   // composantes x, y et z. Chacune de ces composantes ne prenant que trois
   // valeurs -1, 0 et +1. Les positions des 3*3*3 cubes peut donc se dfinir
   // par 27 valeurs diffrentes 0  26. Par commodit de codage 2 bits sont
   // rservs pour chaque composantes et le vecteur complet utilise 6 bits.
   // Pour completer cette definition d'un cube elementaire, son orientation
   // est ajoutee, definie au quart de tour par 24 valeurs differentes de 1
   // a 24 (typedef BYTE t_Matrix).

   t_Mask  m_cube[27];  // position et orientation des cubes constituants

   // L'tat du Rubik's Cube est donc dfini par un tableau de 27 valeurs
   // indiquant la position et l'orientation des 27 cubes elementaires. Par
   // commodit, ce tableau est ordonn aprs chaque phase de dformation
   // par valeurs croissantes du vecteur de position.
   // Les 27 orientations suffisent pour dcrire un Rubik's Cube si elles
   // sont prsentes dans l'ordre dfini ci-dessus.
   // Le cube 13, central, contient l'orientation de tout le Rubik's Cube.
   // Tous les cubes ont cette mme orientation lorsqu'il est ordonn.
   // Un Rubik's Cube standard peut paraitre ordonn, avec des faces de
   // couleurs uniformes, alors que les cubes de centres de faces sont
   // tourns. Certes, il prsentent la bonne couleur de face, mais les
   // faces latrales ne sont pas bien orientes...

   float   m_dim;       // dimension d'une face en 3D

   CSize   m_screen;    // dimensions fenetre affichage
   CRect   m_box;       // rectangle englobant la figure sur l'ecran

          // au maximum 27*3 faces visibles + 1 face de travail pour Sides::Create()
   BYTE    m_nbSide;    // nombre de faces visibles
   CSide   m_side[81];  // ensemble de faces visibles (activees par Sides::Create())

   CSide   m_si2D[54];  // faces calculee en cube developpe a plat

   t_Matrix m_oldMove;

   void    SZSort(int low, int high);// tri des faces visibles en z
   void    SPSort(int low, int high);// tri des faces en position cube

   //     variables fournies par l'application :
   int    *m_pYObs;     // distance observateur
   const CMatrix*m_pMro;// pt matrice passage repere RubixCube vers Observateur
   CMatrix*m_pMmr;      // pt matrice passage repere partie Mobile vers RubixCube

friend class CMove;
};

/////////////////////////////////////////////////////////////////////////////

extern void InitColorsRotateTable();

extern t_Matrix GetOrientationSide(BYTE noSide);

extern BYTE GetCubeSideColor(t_Matrix or, BYTE face);

// Gives the orientation of the cube on the basis of colors of two faces
extern t_Matrix GetCubeOrientation(t_GROUP colors, t_GROUP noSide,
                                               t_Matrix def = MatrixUnit);
extern t_Matrix GetCubeOrientation(const BYTE facets[54]);

///////////////////////////////////////////////////////////////////////////////

#endif // !defined(AFX_CUBES_H__7752AAB1_D4AC_11D3_A5F0_AA6FFDA5D841__INCLUDED_)
