// Matrix.cpp: implementation of the CMatrix class.
//
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Matrix.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//=========================================================================//
static void mUnPack(t_Matrix pack, char m[3][3]);
static bool initMatrixTables();
static bool ok = initMatrixTables();
static const float PI = 3.14159265359f;
//=========================================================================//

static BYTE tabCmdRot[32] = {0x55, };
static BYTE tabQrTurn[64] = {0, };

static t_Matrix invertTbl[25] = {0, };
static t_Matrix producTbl[25][25] = {0,};
static t_Mask tabVectRot[27][25] = {0, };

// rotation plan in the matrix       0,x    1,y    2,z
static const BYTE rotPlan[3][2] = { {1,2}, {2,0}, {0,1} };
// for mUnPack functions
static const WORD matAddr[8] = { 0x012,0x021,0x120,0x102,0x201,0x210,0,0 };

/////////////////////////////////////////////////////////////////////////////
//           This functions are used only to generates a tables            //
//=========================================================================//
//                                                                         //
//           Convert a unit char matrix 3 * 3 to a number 1 to 24          //
//                                                                         //
BYTE mPack(const char m[3][3])                                             //
//=========================================================================//
{
   t_Matrix pack = 0;

   // The contents of the third column is ignored, it will be reconstituted
   // in the inverse operation with the first two columns values.
   // The code is the form (Nab) + 1 where N is 0 to 5 to define the cases of
   // ids in the two first column, a the sign of id the second column and b
   // the sign of id the first column.
   if (m[0][0] != 0) {
      if (m[0][0] < 0) pack = 1;

      if (m[1][1] != 0) {                 // x 0 0
         if (m[1][1] < 0) pack |= 2;      // 0 x 0 => 0 1 2
         pack += 1 + 0 * 4;               // 0 0 x
      } else {
         if (m[1][2] < 0) pack |= 2;      // x 0 0
         pack += 1 + 1 * 4;               // 0 0 x => 0 2 1
      }                                   // 0 x 0
   } else if (m[0][1] != 0) {
      if (m[0][1] < 0) pack = 1;

      if (m[1][2] != 0) {                 // 0 0 x
         if (m[1][2] < 0) pack |= 2;      // x 0 0 => 1 2 0
         pack += 1 + 2 * 4;               // 0 x 0
      } else {
         if (m[1][0] < 0) pack |= 2;      // 0 x 0
         pack += 1 + 3 * 4;               // x 0 0 => 1 0 2
      }                                   // 0 0 x
   } else if (m[0][2] != 0) {
      if (m[0][2] < 0) pack = 1; 

      if (m[1][0] != 0) {                 // 0 x 0
         if (m[1][0] < 0) pack |= 2;      // 0 0 x => 2 0 2
         pack += 1 + 4 * 4;               // x 0 0
      } else {
         if (m[1][1] < 0) pack |= 2;      // 0 0 x
         pack += 1 + 5 * 4;               // 0 x 0 => 2 1 0
      }                                   // x 0 0
   }
   return pack;
}

//=========================================================================//
//                                                                         //
//             Convert a number 1 to 24 in a char matrix 3 * 3             // 
//                                                                         //
static void mUnPack(t_Matrix pack, char m[3][3])                           //
//=========================================================================//
{
   BYTE p = (--pack >> 2) & 7;
   WORD a = matAddr[p];

   memset(m, 0, 9);

   m[0][a >> 8] =             (pack & 1)? -1: 1;
   m[1][(a >> 4)& 3] =        (pack & 2)? -1: 1;
   m[2][a & 3] = ((pack^(pack>>1)^p)& 1)? -1: 1;
}

//=========================================================================//
//                                                                         //
//             Convert a number 1 to 24 in a float matrix 3 * 3            //
//                                                                         //
static void mUnPack(t_Matrix pack, float mod, float m[3][3])               //
//=========================================================================//
{
   BYTE p = (--pack >> 2) & 7;
   WORD a = matAddr[p];

   m[0][0]= m[0][1]= m[0][2]= m[1][0]= m[1][1]= m[1][2]= m[2][0]= m[2][1]=
                                 m[2][2]= .0f;

   m[0][a >> 8] =             (pack & 1)? -mod: mod;
   m[1][(a >> 4)& 3] =        (pack & 2)? -mod: mod;
   m[2][a & 3] = ((pack^(pack>>1)^p)& 1)? -mod: mod;
}
//=========================================================================//

static void mInvert(char m[3][3])
{
   char t;

   // exchange between row and column
   for (int a = 0; a < 3; a++)
      for (int b = 0; b < a; b++)
         t = m[a][b], m[a][b] = m[b][a], m[b][a] = t;
}

static void mProduct(const char Myz[3][3], const char Mxy[3][3],
                              char Mxz[3][3])
{
   for (int i = 0; i < 3; i++)
      for (int j = 0; j < 3; j++)
         Mxz[i][j] = Mxy[i][0] * Myz[0][j] + Mxy[i][1] * Myz[1][j] +
                              Mxy[i][2] * Myz[2][j]; 
}

static void vProduct(const char v[3], const char m[3][3], char vo[3])
{
   char x = v[0], y = v[1], z = v[2];

   for (int i = 0; i < 3; i++)
      vo[i] = x * m[0][i] + y * m[1][i] + z * m[2][i];
}

static void mQRotate(BYTE axe, bool neg, char m[3][3])
{
   BYTE a, b, i;
   char t;

   // calcul du plan de rotation
   if (neg) a = rotPlan[axe][1], b = rotPlan[axe][0];
   else     a = rotPlan[axe][0], b = rotPlan[axe][1];

   // rotation
   for (i = 0; i < 3; i++)
      t = m[i][a], m[i][a] = m[i][b], m[i][b] = -t;
}

static CString mDumpToString(const char m[3][3])
{
   CString out;

   out.Format("\n%6d%3d%3d\n%6d%3d%3d\n%6d%3d%3d\n", m[0][0],m[0][1],m[0][2],
                     m[1][0],m[1][1],m[1][2], m[2][0],m[2][1],m[2][2]);
   return out;
}

static char entryCnt(BYTE b)  // count the number of command
{                             //  (an order holds a 2-bit field)
   char n;

   if (b == 0xff) n = 16;
   else for (n = 0; b ; b >>= 2)
      if (b & 3) n += ((b & 3) == 2)? 3: 2;  // promotes the 90 degrees

   return n;
}

//=========================================================================//
//   Tables generating for rotations of quarter turns on all three axes.   //
//=========================================================================//
//                                                                         //
// To determine the coefficients of different tables, 3 * 3 matrices are   //
// used whose coefficients can take only the values +1, 0 or -1.           //
// These matrices can represent 24 possible orientations numbered from 1   //
// to 24. The first value 1 represents an unit orientation.                //

static bool initMatrixTables()
{
   char m1[3][3], m2[3][3], m3[3][3];
   char v1[3], v2[3];
   t_Matrix a, u, v;
   BYTE c, x, y, z;
   DWORD mask = 0;

   // Make rotation tables

   // initial value of tables
   for (x = 0; x < 32; x++) tabCmdRot[x] = 0xFF;

   mUnPack(MatrixUnit, m1);
   // for the 4 directions on the three axes (64 cases)
   for (x = 0; x < 4; x++) {
      memcpy(m2, m1, 9);
      for (y = 0; y < 4; y++) {
         memcpy(m3, m2, 9);
         for (z = 0; z < 4; z++) {
            // 64 cases generates only 24 different orientation values

            a = mPack(m3);
            ASSERT(a < 25);
            mask |= 1 << a;

            c = x | (y << 2) | (z << 4);  // => 00zzyyxx
            // choose the shortest order of rotation
            if (entryCnt(c) < entryCnt(tabCmdRot[a])) tabCmdRot[a] = c;
            tabQrTurn[c] = a;

            mQRotate(2, false, m3);       // +z
         }
         mQRotate(1, false, m2);          // +y
      }
      mQRotate(0, false, m1);             // +x
   }
   ASSERT(mask == (1 << 25) - 2);         // test all orientation

   // make Product and invert tables

   invertTbl[0] = 0;
   producTbl[0][0] = 0;
   for (u = MatrixUnit; u < 25; u++) {
      DWORD maskp = 0;

      // product table
      producTbl[u][0] = producTbl[0][u] = 0;
      mUnPack(u, m1);
      for (v = MatrixUnit; v < 25; v++) {
         mUnPack(v, m2);
         mProduct(m1, m2, m3);
         producTbl[u][v] = a = mPack(m3);
         maskp |= 1 << a;
      }
      ASSERT(maskp == (1 << 25) - 2);

      // invert table
      mInvert(m1);
      invertTbl[u] = v = mPack(m1);
      mask ^= 1 << v;
   }
   ASSERT(mask == 0);

   // make vector product table

   for (x = 0; x < 27; x++) {
      // for all vectors values 3*3*3
      vUnPack(x, v1);

      tabVectRot[x][0] = 0;
      for (u = MatrixUnit; u < 25; u++) {
         // for all orientations
         mUnPack(u, m1);
         vProduct(v1, m1, v2);
         tabVectRot[x][u] = vectPack(v2);
      }
   }
   return true;
}

bool initMatrixTest() { return tabCmdRot[0] != 0x55; }

//=========================================================================//
//                                                                         //
t_Matrix MatrixInvert(t_Matrix mat)                                        //
//=========================================================================//
{
   ASSERT(initMatrixTest());
   ASSERT(mat < 25);
   return invertTbl[mat];
}

//=========================================================================//
//                                                                         //
t_Matrix MatrixProduct(t_Matrix x, t_Matrix y)                             //
//=========================================================================//
{
   ASSERT(initMatrixTest());
   ASSERT(x < 25 && y < 25);
   return producTbl[x][y];
}

//=========================================================================//
//                                                                         //
//       xyz => 00zzyyxx:=> x(0 to 3q), y(0 to 3q), z(0 to 3q)             //
//                                                                         //
t_Matrix MatrixOrient(BYTE xyz)                                            //
//=========================================================================//
{
   ASSERT(initMatrixTest());
   ASSERT(xyz < 64);
   return tabQrTurn[xyz];
}

//=========================================================================//
//                                                                         //
//       return 00zzyyxx:=> x(0 to 3q), y(0 to 3q), z(0 to 3q)             //
//                                                                         //
BYTE toQarterTurn(t_Matrix mat)                                            //
//=========================================================================//
{
   ASSERT(initMatrixTest());
   ASSERT(mat > 0 && mat < 25);
   return tabCmdRot[mat];
}

//=========================================================================//
//                                                                         //
t_Mask VectorOrientation(t_Mask vect, t_Matrix mat)
//=========================================================================//
{
   ASSERT(initMatrixTest());
   ASSERT(mat > 0 && mat < 25);
   return tabVectRot[cubeNo(vect)][mat];
}

/////////////////////////////////////////////////////////////////////////////

//=========================================================================//
//                                                                         //
//               Calcule le produit de deux matrices 3 * 3                 //
//                                                                         //
static void MatrixProduct(const float Myz[3][3], const float Mxy[3][3],    //
                          float Mxz[3][3])                     // resultat //
//=========================================================================//
{
   for (int i = 0; i < 3; i++)
      for (int j = 0; j < 3; j++)
         Mxz[i][j] = Mxy[i][0] * Myz[0][j] +  
                       Mxy[i][1] * Myz[1][j] +
                         Mxy[i][2] * Myz[2][j]; 
}

//=========================================================================//
//                                                                         //
//          Calcule le produit inverse de deux matrices 3 * 3              //
//                                                                         //
static void MatrixIProduct(const float Myz[3][3],const float Mxy[3][3],    //
                           float Mxz[3][3])                    // resultat //
//=========================================================================//
{
   for (int i = 0; i < 3; i++)
      for (int j = 0; j < 3; j++)
         Mxz[i][j] = Mxy[0][i] * Myz[j][0] +  
                       Mxy[1][i] * Myz[j][1] +
                         Mxy[2][i] * Myz[j][2]; 
}

//=========================================================================//
//                                                                         //
//                  Translate a cmd to rotation angle                      //
//                                                                         //
t_Matrix CmdToRotate(t_Command cmd)                                        //
//=========================================================================//
{
   t_Matrix rot = MatrixUnit;

   if (cmd & 0x80) {
      // rotations multiples du Rubik's Cube complet

      if (cmd & 0x40) {
         // rotation negative suivant les 3 axes

         for (char axe = 2, or; axe >= 0; axe--)
            if (or = ((cmd >> (axe * 2))& 3)) //Rotate90Req(4 - or + axe * 4);
               rot = MatrixProduct(rot, MatrixOrient(4 - or + axe * 4));

      } else {
         // rotation positive suivant les 3 axes

         for (char axe = 0, or; axe < 3; axe++)
            if (or = ((cmd >> (axe * 2))& 3)) //Rotate90Req(or + axe * 4);
               rot = MatrixProduct(rot, MatrixOrient(or + axe * 4));
      }
   } else if (cmd >= 6) {
      t_Mask move = CmdCubeSet(cmd);
      short axe = GetAxe(move) * 2;
      BYTE data = ((cmd & 0x1F)/ 6 == 2)? 2: (move & 0x8000)? 3: 1;

      rot = MatrixOrient(data << axe);
   }
   return rot;
}

/////////////////////////////////////////////////////////////////////////////
//                            CMatrix Class                                //
/////////////////////////////////////////////////////////////////////////////

//=========================================================================//
//                                                                         //
CMatrix::CMatrix(t_Matrix mat, float size)                                  //
//=========================================================================//
{
   mUnPack(mat, size, m_d);
}

//=========================================================================//
//                                                                         //
CMatrix::CMatrix(t_Matrix mat, const CMatrix *pMatrix, float size)         //
//=========================================================================//
{
   float m[3][3];

   mUnPack(mat, size, m);

   if (pMatrix) MatrixProduct(pMatrix->m_d, m, m_d);
   else         memcpy(m_d, m, sizeof(m));
}

//=========================================================================//
//                                                                         //
CMatrix::CMatrix(const CMatrix *pMyz, const CMatrix *pMxy)                 //
//=========================================================================//
{
   MatrixProduct(pMyz->m_d, pMxy->m_d, m_d);
}

//=========================================================================//
//                                                                         //
//              Compacte le contenu de la matrice (unitaire)               //
//                                                                         //
t_Matrix CMatrix::Pack() const                                             //
//=========================================================================//
{
   char m[3][3], s;
   int i, j, k;

   // Convert the matrix into unitary matrix
   memset(m, 0, 9);
   for (i = 0; i < 3; i++) {
      float max = .0f, tmp;

      // detects the max of the column and its sign
      for (k = -1, j = 0; j < 3; j++)
         if ((tmp = m_d[i][j]) > max) max = tmp, k = j, s = 1;
         else if (-tmp > max) max = -tmp, k = j, s = -1;

      if (k < 0) return 0;
      m[i][k] = s;
   }

   return mPack(m);
}

//=========================================================================//
//                                                                         //
CMatrix CMatrix::Invert()                                                  //
//=========================================================================//
{
   float t;

   // echange entre ligne et colonne
   for (int a = 0; a < 3; a++)
      for (int b = 0; b < a; b++)
         t = m_d[a][b], m_d[a][b] = m_d[b][a], m_d[b][a] = t;

   return *this;
}

//=========================================================================//
//                                                                         //
//            Assure une rotatation progressive suivant un axe             //
//                                                                         //
void CMatrix::Rotate(short axe,     // axe rotation 0=> x, 1=> y, 2=> z    //
                     short nb)      // nombre d'increments                 //
//=========================================================================//
{
   float t, inc = PI * 0.0005f;// increment sin(a) equivaut a l'angle pour les 
   int a, b;                   // petits angles on peut confondre angle et sin

   // calcul du plan rotation
   if (nb < 0) a = rotPlan[axe][1], b = rotPlan[axe][0], nb = -nb;
   else        a = rotPlan[axe][0], b = rotPlan[axe][1];

   while (--nb >= 0) {           // rotation de la matrice dans plan (a, b)
      t = m_d[0][a], m_d[0][a] += inc * m_d[0][b], m_d[0][b] -= inc * t;
      t = m_d[1][a], m_d[1][a] += inc * m_d[1][b], m_d[1][b] -= inc * t;
      t = m_d[2][a], m_d[2][a] += inc * m_d[2][b], m_d[2][b] -= inc * t;
   }
}

//=========================================================================//
//                                                                         //
// Rotation de quart de tour, dans le sens desire et suivant un axe donne  //
//                                                                         //
CMatrix CMatrix::Rotate90(short axe,   // 0=> x, 1=> y, 2=> z              //
                          bool neg)    // sens                             //
//=========================================================================//
{
   int a, b;
   float t;

   // calcul du plan de rotation
   if (neg) a = rotPlan[axe][1], b = rotPlan[axe][0];
   else     a = rotPlan[axe][0], b = rotPlan[axe][1];

   // rotation
   t = m_d[0][a], m_d[0][a] = m_d[0][b], m_d[0][b] = -t;
   t = m_d[1][a], m_d[1][a] = m_d[1][b], m_d[1][b] = -t;
   t = m_d[2][a], m_d[2][a] = m_d[2][b], m_d[2][b] = -t;

   return *this;
}

//=========================================================================//
//                                                                         //
void CMatrix::Rotate90Req(BYTE rq)// bits0,1: 0=> 0, 1=> 90, 2=> 180, 3=>270
                                 // bits2,3 : axe 0=> x, 1=> y, 2=> z
//=========================================================================//
{
   if (rq & 3) {
      // angle de rotation different de 0 degre
      short axe = (rq >> 2) & 3;

      if (rq & 1) Rotate90(axe, (rq & 2)== 0);     // plus ou moins 90 degres
      else if (rq & 2) Rotate90(axe, true), Rotate90(axe, true);// 180 degres
   }
}

//=========================================================================//
//                                                                         //
void CMatrix::Rotate90Mask(t_Mask mask, bool dbl)
//=========================================================================//
{
   if (mask & 0x003F) {
      short axe = GetAxe(mask);
      bool neg = (mask & 0x8000) != 0;

      Rotate90(axe, neg);
      if (dbl) Rotate90(axe, neg);
   }
}

//=========================================================================//
//                                                                         //
//       Assure la rotation de la matrice correspondant a la commande      //
//                                                                         //
CMatrix CMatrix::CmdSet(t_Command cmd)                                     //
//=========================================================================//
{
   if (cmd & 0x80) {
      // rotations multiples du Rubik's Cube complet

      if (cmd & 0x40) {
         // rotation negative suivant les 3 axes

         for (char axe = 2, or; axe >= 0; axe--)
            if (or = ((cmd >> (axe * 2))& 3)) Rotate90Req(4 - or + axe * 4);
      } else {
         // rotation positive suivant les 3 axes

         for (char axe = 0, or; axe < 3; axe++)
            if (or = ((cmd >> (axe * 2))& 3)) Rotate90Req(or + axe * 4);
      }
   } else if (cmd >= 6) {
      t_Mask move = CmdCubeSet(cmd);

      Set(1.0f);
      Rotate90Mask(move, (cmd & 0x1F) / 6 == 2);
      // ordre de rotations de faces
//      BYTE face = (cmd & 0x1F) % 6, or = (cmd & 0x1F) / 6, axe = face / 2;

      // axe x: R - L, axe y: F - B, axe z: U - D
//      if (face & 1) or = (4 - or) & 3;
//      Rotate90Req(or + axe * 4);
   }
   return *this;
}

//=========================================================================//
//                                                                         //
void CMatrix::RotateCmd(t_Mask cmd,    // ordre de rotation (sens, axe)    //
                        short nb)      // amplitude de la rotation         //
//                                                 (en nbre increments)    //
//=========================================================================//
{
   if (cmd & 0x003F) {
      if (cmd & 0x8000) nb = -nb;
      Rotate(GetAxe(cmd), nb);
   }
}

//=========================================================================//
//                                                                         //
//        conversion d'un deplacement souris en rotation du cube           //
//                                                                         //
void CMatrix::PosRotate(int x, int y)                                      //
//=========================================================================//
{         // 
   int nx = x, ny = y;
   short axe = 2;

   if (x < 0) x = -x;
   if (y < 0) y = -y;
   if (x > y) axe = 0; else nx = ny;

   Rotate(axe, nx * 5);
}

//=========================================================================//
//                                                                         //
DWORD CMatrix::ToCommand(t_Matrix or, bool neg)                            //
//=========================================================================//
{
   BYTE rot = toQarterTurn(or), ng = (neg)? 1: 3;
   DWORD out = 0;

   ASSERT(rot != 0xFF);

   if (rot == 0) return out;       // commande nulle

   for (BYTE axe = 1; rot != 0; axe++, rot >>= 2) {
      if (rot & 3) {
         out = (out << 8) | (axe | 0x0C);
         if ((rot & 3) == 2) out = (out << 8) | BYTE(out);// double commande
         else if ((rot & 3) == ng) out |= 0x80;          // commande negative
      }
   }

   if (neg) return out;


   DWORD out2 = 0;      // inverse l'ordre des commandes
   while (out) out2 = (out2 << 8) | BYTE(out), out >>= 8;

   return out2;
}

//=========================================================================//
//                                                                         //
//                debug : matrix display orientation or position           //
//                                                                         //
CString CMatrix::DumpMat(t_Matrix mat, const CString &text)                //
//=========================================================================//
{
   char m[3][3];
   CString msg;

   mUnPack(mat, m);

   msg.Format( "%s fr=%2d  %3d%3d%3d  %3d%3d%3d  %3d%3d%3d ", text, mat,
                     m[0][0], m[0][1], m[0][2], m[1][0], m[1][1],
                     m[1][2], m[2][0], m[2][1], m[2][2] );
   return msg;
}

//=========================================================================//
//                                                                         //
CString CMatrix::Dump(short mode) const                                    //
//=========================================================================//
{
   CString msg, tmp;
   if (mode & 1) {
      for (int i = 0; i < 3; i++) {
         tmp.Format(" %7.4f %7.4f %7.4f\n", m_d[i][0], m_d[i][1], m_d[i][2]);
         msg += tmp;
      }
   }
   if (mode & 2) {
      msg += DumpMat(Pack(), " t_Matrix ");
      msg += "\n";
   }
   return msg;
}

/////////////////////////////////////////////////////////////////////////////
