// Matrix.h: interface for the CMatrix class.
//
/////////////////////////////////////////////////////////////////////////////

#if !defined(AFX_MATRIX_H__C400A8C2_A8CE_4E93_B00C_E83F0411FFAC__INCLUDED_)
#define AFX_MATRIX_H__C400A8C2_A8CE_4E93_B00C_E83F0411FFAC__INCLUDED_

#include "CubeUtilities.h"

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

/////////////////////////////////////////////////////////////////////////////
typedef BYTE t_Matrix;  // Value ranging from 1 to 24 inclusive, representing
                        // the 24 possible orientations of a cube in quarter
                        // turn (modulo 90 degrees in x, y, z axis).

const t_Matrix MatrixUnit = 1;   // initial orientation (x = 0, y = 0, z = 0)
const t_Matrix MatrixInvalid = 0;// invalid value

extern t_Matrix MatrixInvert(t_Matrix mat);
extern t_Matrix MatrixProduct(t_Matrix x, t_Matrix y);

extern t_Matrix MatrixOrientation(BYTE axe, bool neg);

// 00zzyyxx:=> 0 to 3q
extern t_Matrix MatrixOrient(BYTE xyz);
extern BYTE toQarterTurn(t_Matrix mat);

extern t_Mask VectorOrientation(t_Mask v, t_Matrix or);

extern t_Matrix mPack(const char mat[3][3]);

extern t_Matrix CmdToRotate(t_Command cmd);


extern bool initMatrixTest();


/////////////////////////////////////////////////////////////////////////////
class CMatrix           // matrice 3 * 3
/////////////////////////////////////////////////////////////////////////////
{//  Une matrice permettant le passage du repere x au repere y est appelee Mxy
public:
   CMatrix() { Set(1.0f); }

   CMatrix(const float m[3][3]) { memcpy(m_d, m, sizeof(m_d)); }

   CMatrix(t_Matrix mat, float size = 1.0f);

   CMatrix(t_Matrix mat, const CMatrix *matrix, float size = 1.0f);

   CMatrix(const CMatrix *pMyz, const CMatrix *pMxy); // genere Mxz

   virtual ~CMatrix() {}

   ////////////////////////////////////////////////////////////////////////

   inline void Set(float fl = 1.0f) {  m_d[0][0]=m_d[1][1]=m_d[2][2] = fl;
      m_d[1][0]=m_d[2][0]=m_d[0][1]= m_d[2][1]=m_d[0][2]=m_d[1][2] = 0.0f; }

//   static t_Matrix matPack(const char mat[3][3]);

   void    Rotate(short axe, short nb);
   CMatrix Rotate90(short axe, bool neg);
   CMatrix Invert();

   void Rotate90Mask(t_Mask mask, bool dbl);

   void    Rotate90Req(BYTE rq);
   inline static BYTE InvertReq(BYTE rq) { return (rq & 1)? rq ^ 2: rq; }

   void    RotateCmd(t_Mask cmd, short nb);

   void    PosRotate(int x, int y);

//   static BYTE CmdRot(t_Matrix mat);
   inline t_Command CmdGetRotation(bool invert = false) const
      { t_Matrix m = Pack(); return (m!=1)? toQarterTurn(m)|(invert?0xC0:0x80): 0; }

   static DWORD ToCommand(t_Matrix mat, bool neg);
   inline DWORD ToCommand(bool neg) const
      { return ToCommand(Pack(), neg); }

   // defines a unitary matrix has around 90 degrees
   t_Matrix Pack() const;

   CMatrix CmdSet(t_Command cmd);

   static CString DumpMat(t_Matrix mat, const CString &text);
   CString Dump(short mode = 3) const;

   inline void operator=(const CMatrix &mat) { memcpy(m_d, mat.m_d, sizeof(m_d)); }
   inline void operator=(t_Matrix mat) { *this = CMatrix(mat); }

   inline bool operator==(t_Matrix or) const { return Pack() == or; }
   inline bool operator!=(t_Matrix or) const { return Pack() != or; }

protected:
   float	m_d[3][3];     // matrix coefficients

   friend class CCubes;
   friend class CVertex;
   friend class CCubeVertexList;
};

#endif // !defined(AFX_MATRIX_H__C400A8C2_A8CE_4E93_B00C_E83F0411FFAC__INCLUDED_)
