// Sides.cpp: implementation of the CVertex, CCubeVertexList, CSide classes.
//
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Cubes.h"
#include "Sides.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

/////////////////////////////////////////////////////////////////////////////
//                 class CSide (facet of Rubik's Cube)                     //
/////////////////////////////////////////////////////////////////////////////
// name of faces (central facet of each faces of the Rubik's Cube)
static TCHAR pid[] = { "    D     F L R B     U    " };
//=========================================================================//
// translate integer, float and reverse

static float gVex(t_Mask mask, float module)
   { mask &= 3; ASSERT(mask != 0);
      return (mask == 2)? 0.0f: (mask == 3)? module: -module; }

static WORD gfPos(float x) { return ((x < -0.1f)? 1: ((x > 0.1f)? 3: 2)); }


//=========================================================================//
//                                                                         //
//       Initializes a face of the cube, determines if it is visible       //
//                                                                         //
BOOL CSide::Create(CCubeVertexList &vrtx, // position of the vertices      //
                   BYTE noSide,           // facet number, color           //
                   t_Mask cube,           // orientation & cube position   //
                   BYTE noCube,           // cune no (initial position)    //
                   CPoint org)            // origin of display picture     //
//=========================================================================//
{  // Creation of a face, return TRUE if the face is turned towards the observer
   int x1, x2, y1, y2, x, y, v;

   // make the polygon defining the facet
   vrtx.Side(noSide, this, org);

   // tests whether the facet is turned towards the observer
   int dx1 = m_poly[0].x - m_poly[1].x, dy1 = m_poly[0].y - m_poly[1].y;
   int dx2 = m_poly[0].x - m_poly[2].x, dy2 = m_poly[0].y - m_poly[2].y;
   if (dx1*dy2 > dx2*dy1) {               // it is visible

      // calculate the bounding rectangle of the facet (box) 
      x1 = x2 = m_poly[0].x, y1 = y2 = m_poly[0].y;
      for (v = 1; v < 4; v++) {
         if ((x = m_poly[v].x) < x1) x1 = x; if (x > x2) x2 = x;
         if ((y = m_poly[v].y) < y1) y1 = y; if (y > y2) y2 = y;
      }
      m_box.SetRect(x1, y1, x2+1, y2+1);  // box

      // facet attributs
      m_cube   = cube;        // position and orientation of the cube
      m_noSide = noSide;      // face number => facet color
      m_noDscr = 0xFF;        // not used in 3D
      m_pSide  = this;        // initialize the classification of the facets
      return TRUE;                     // visible facet
   }
   return FALSE;                       // not visible facet
}

//=========================================================================//
//                                                                         //
//      Calculate an elementary face in flat developed representation      //
//                                                                         //
BOOL CSide::Create2D(CMatrix &meo,  // viever direction matrix             //
                     t_Mask cube,   // elementary cube                     //
                     BYTE noDscr,   // entry no of descripteur, color      //
                     CPoint org,    // offset position to draw             //
                     float dim,     // size of an facet                    //
                     float ds)      // demi cote d'une face en 2D          //
//=========================================================================//
{
   CCubeVertexList vrtx(cube, meo, dim, ds); // 8 sommets du cube elementaire

   // must choose the right face whose all vertices z is positive (front),
   // which allow people to know the color

   for (BYTE noSide = 0; noSide < 6; noSide++) {
      // eliminates the faces rear and side
      if (!vrtx.IsZPos(noSide)) continue;

      // the 4 vertex of this facet
      vrtx.Side(noSide, this, org);

      m_box = CRect(m_poly[0], m_poly[2]); 
      m_box.NormalizeRect();  // box
      m_z     = 0;
      m_cube  = cube;
      m_noSide= noSide;       // facet color 0..5
      m_noDscr= noDscr;
      m_pSide = this;

      return TRUE;
   }
   return FALSE;
}

//=========================================================================//
//                                                                         //
//                            Draw the facet                               //
//                                                                         //
void CSide::Draw(CDC *pDC)                                                 //
//=========================================================================//
{
   pDC->SelectObject(GetBrush(m_noSide));
   pDC->Polygon(m_poly, 4);
}

//=========================================================================//
//                                                                         //
//                            Draw the facet id                            //
//                                                                         //
void CSide::DrawId(CDC *pDC)
//=========================================================================//
{
   TCHAR id = pid[cubeNo(m_cube)];
   if (id != TCHAR(' ')) {
      COLORREF oldColor = pDC->SetTextColor(RGBTextColor(m_noSide));
      pDC->DrawText(&id, 1, &m_box, DT_CENTER|DT_VCENTER|DT_SINGLELINE);
      pDC->SetTextColor(oldColor);
   }
}

//=========================================================================//
//                                                                         //
//                Tests if the position is inside the facet                //
//                                                                         //
BOOL CSide::PtInSide(const CPoint &pos) const                              //
//=========================================================================//
{
   if (m_box.PtInRect(pos)) {
      // the point is inside the box
      CPoint pV1 = m_poly[3] - pos, pV2 = m_poly[0] - pos;
      int s0 = pV1.x * pV2.y - pV1.y * pV2.x;

      // compare the sign of the surfaces of triangles formed by point and
      // the sides of the face
      for (BYTE i = 1; i < 4; i++) {
         pV1 = pV2, pV2 = m_poly[i] - pos;
         int sn = pV1.x * pV2.y - pV1.y * pV2.x;
         if ((sn ^ s0) < 0)
            return FALSE;
      }
      return TRUE;
   }
   return FALSE;
}

//=========================================================================//
//                                                                         //
//           Compare the positions of the sides in held flat               //
//                                                                         //
int CSide::Compare(CSide *pSide) const                                     //
//=========================================================================//
{
   const CRect &ra = m_box, &rb = pSide->m_box;

   return (ra.top == rb.top)? ra.left - rb.left: ra.top - rb.top;
}

//=========================================================================//
//                                                                         //
//       Calculating the bounding rectangle (box) of a list of faces       //
//                                                                         //
CRect CSide::GetRect(const CSide pSide[], int nbSide)                      //
//=========================================================================//
{
   int x1 = pSide->m_poly[0].x, x2 = x1, y1 = pSide->m_poly[0].y, y2 = y1;

   for (int n = 0; n < nbSide && pSide->m_pSide; n++, pSide++) {
      const CPoint *pt = pSide->m_poly;

      for (int i = 0; i < 4; i++, pt++) {
         // calculating the bounding rectangle (box) of the facet
         if (pt->x < x1) x1 = pt->x; if (pt->x > x2) x2 = pt->x;
         if (pt->y < y1) y1 = pt->y; if (pt->y > y2) y2 = pt->y;
      }
   }

   return CRect(x1, y1, x2, y2);
}

//=========================================================================//
//                                                                         //
//          Give the message corresponding to the face (debug)             //
//                                                                         //
char *CSide::GetMsg(short mode,                 // mode of display         //
                    const CCubes *pCube) const                             //
//=========================================================================//
{           // retourne les caracteristiques de la face
   DWORD centers = pCube? pCube->CenterSidesRotation(): 0;
   BYTE v = BYTE((m_cube >> 9) & 0x003F), n1, n2, n3;
   BYTE k = cubeNo(m_cube);
   TCHAR faceLeter = pid[k];
   static CString msg, tmp;

   switch (mode % MODE_TOOLTIP) {
   case 0:
      tmp = faceLeter != TCHAR(' ')? GetMsgCenter(centers, faceLeter): "";
      msg.Format("Cube n%d, %s code %04X ",
         k+1, tmp, m_cube);
      break;
   case 1:
      n1 = k % 3, n2 = (k % 9)/ 3, n3 = k / 9;
      msg.Format("Cube [%d,%d,%d], Emplacement [%d,%d,%d], %c code %04X ",
         n1, n2, n3, v & 3, (v >> 2) & 3, (v >> 4) & 3, pid[k], m_cube);
      break;
   case 2:
      tmp.Format("Code cube_%d : ", k+1);
      msg = CMatrix::DumpMat(t_Matrix(m_cube), tmp);
      break;
   case 3:
      {
         CMatrix f(GetOrientationSide(m_noSide));
         CMatrix c(t_Matrix(m_cube), &f);
         tmp.Format("Orientation face_%d : ", m_noSide);
         msg = CMatrix::DumpMat(c.Pack(), tmp);
 
      }
      break;
   case 4:
      msg.Format("pos=(%d,%d) sid=%d d=%d", 0,0, 0, 0);
      break;
   }

   return msg.GetBuffer(0);
}

//=========================================================================//
//                                                                         //
char *CSide::Mode(short mode)                                              //
//=========================================================================//
{           // retourne le mode d'affichage de la face 
   static char *modeMsg[MODE_TOOLTIP] = {
      "Affiche nposition (1..27)",  // 0
      "Affiche [planx,plany,planz]", // 1
      "Position Orientation cube",   // 2
      "Matrice orientation face",    // 3
      "Position, Distance"};         // 4

   return modeMsg[mode % MODE_TOOLTIP];
}


/////////////////////////////////////////////////////////////////////////////
//                    CVertex Class  (position in space 3D)                //
/////////////////////////////////////////////////////////////////////////////

//=========================================================================//
//                                                                         //
CVertex::CVertex(t_Mask pos,           // position                         //
                  const CMatrix *pMat, // orientation matrix               //
                  float mod)           // module of face size              //
//                                                                         //
//=========================================================================//
{
   float x = gVex(pos>>=9, mod), y = gVex(pos>>=2, mod), z = gVex(pos>>=2, mod);

   m_x = x * pMat->m_d[0][0] + y * pMat->m_d[1][0] + z * pMat->m_d[2][0];
   m_y = x * pMat->m_d[0][1] + y * pMat->m_d[1][1] + z * pMat->m_d[2][1];
   m_z = x * pMat->m_d[0][2] + y * pMat->m_d[1][2] + z * pMat->m_d[2][2];
}

/////////////////////////////////////////////////////////////////////////////
//=========================================================================//
//                                                                         //
t_Mask CVertex::Pack()           // compact an unit vector                 //
//                                                                         //
//=========================================================================//
{
   return (gfPos(m_x) << 9) | (gfPos(m_y) << 11) | (gfPos(m_z) << 13);
}

/////////////////////////////////////////////////////////////////////////////
//              CCubeVertexList Class  (8 vertices of a cube)              //
/////////////////////////////////////////////////////////////////////////////

float CCubeVertexList::m_ds3D = 0.42f; // demi cote d'une face 3D
//float CCubeVertexList::m_ds2D = 0.46f; // demi cote d'une face en 2D
//=========================================================================//
//                                                                         //
//          Calculate the 8 vertices of a cube in perspective 2D           //
//                                                                         //
CCubeVertexList::CCubeVertexList(                                          //
                        t_Mask cube,   // cube elementaire                 //
                        const CMatrix *pMeo, // matrice d'orientation du cube    //
                        float dim,     // module de construction du cube   //
                        float size,    // dimension du RubixCube affiche   //
                        float yPos)    // distance observateur             //
//=========================================================================//
{
   float x, y, z;
   CMatrix Mco(t_Matrix(cube), pMeo, dim);// orientation and size
   CVertex pos(cube, pMeo, dim);          // position of Rubik's Cube
   float x0 = pos.m_x, y0 = pos.m_y + yPos, z0 = pos.m_z;
   float u = m_ds3D, v = u, w = u;
   t_VERTEX *pV = m_vertex;

   // calculate the 8 vertices of the cube by symmetry 2-2
   for (short i = 0; i < 2; ++i, u = -u) {
      for (short j = 0; j < 2; ++j, v = -v) {
         for (short k = 0; k < 2; ++k, w = -w, pV++) {
            // passing from reference of the cubic to reference observer
            x = x0 + Mco.m_d[0][0]*u + Mco.m_d[1][0]*v + Mco.m_d[2][0]*w,
            z = y0 + Mco.m_d[0][1]*u + Mco.m_d[1][1]*v + Mco.m_d[2][1]*w,
            y = z0 + Mco.m_d[0][2]*u + Mco.m_d[1][2]*v + Mco.m_d[2][2]*w;
            // perspective translation to reference screen 2D (simplified)
            if (z == 0.0) z = 0.0001f;    // to avoid the crash if 0
            pV->v.x = int( x / z * size);
            pV->v.y = int(-y / z * size);
            pV->z   = int(-z);// observer distance for Z buffer management
         }
      }
   }
}

//=========================================================================//
//                                                                         //
//                Calculate the 8 vertices of a cube in 2D                 //
//                                                                         //
CCubeVertexList::CCubeVertexList(                                          //
                     t_Mask cube,         // cube elementaire              //
                     const CMatrix &Meo,  // matrice d'orientation du cube //
                     float dim,           // dimension face elementaire    //
                     float ds)            // demi cote d'une face en 2D    //
//=========================================================================//
{
   float u = ds, v = u, w = u;
   CMatrix Mco(t_Matrix(cube), &Meo, dim);// orientation and size
   CVertex pos(cube, &Meo, dim);          // cube position
   t_VERTEX *pV = m_vertex;               // 8 vertices of the cube


   // calculate the 8 vertices of the cube by symmetry 2-2
   for (BYTE i = 0; i < 2; ++i, u = -u) {
      for (BYTE j = 0; j < 2; ++j, v = -v) {
         for (BYTE k = 0; k < 2; ++k, w = -w, pV++) {
            // passing from reference of the cubic to reference observer
            // (expanded flat)
            pV->v.x = int(pos.m_x+Mco.m_d[0][0]*u+Mco.m_d[1][0]*v+Mco.m_d[2][0]*w);
            pV->z =  -int(        Mco.m_d[0][1]*u+Mco.m_d[1][1]*v+Mco.m_d[2][1]*w);
            pV->v.y =-int(pos.m_z+Mco.m_d[0][2]*u+Mco.m_d[1][2]*v+Mco.m_d[2][2]*w);
         }
      }
   }
}

//=========================================================================//

// gives the number of vertices of the face        4 ------- 0
static const short cubeFaces[6][4] = { //          | \    up  \
// 0 Right    1 Left                               |   \     |  \ 
   {0,1,3,2}, {6,7,5,4},   //  Show side           |    6 ------- 2 
// 2 Back     3 Front          vertex            left   |    |    |right
   {4,5,1,0}, {2,3,7,6},   //  rotate CW,          5  - | -  1    |
// 4 Up       5 Down           hide CCW              \  |     \   |
   {0,2,6,4}, {1,5,7,3}    //                          \| down  \ |
};                         //                           7 ------- 3

//=========================================================================//
//                                                                         //
bool CCubeVertexList::IsZPos(BYTE noSide) const                            //
//=========================================================================//
{
   return ( m_vertex[cubeFaces[noSide][0]].z > 0
         && m_vertex[cubeFaces[noSide][1]].z > 0
         && m_vertex[cubeFaces[noSide][2]].z > 0
         && m_vertex[cubeFaces[noSide][3]].z > 0 );
}

//=========================================================================//
//                                                                         //
//                   Isolates the 4 vertices of the face                   //
//                                                                         //
void CCubeVertexList::Side(BYTE noSide,      // no of the face 0..5        //
                           CSide *pSide,     // face to update             //
                           const CPoint &ofs // offset to add at x and y   //
                           ) const                                         //
//=========================================================================//
{
   int v, z;

   // generate the polygon defining the face
   for (v = 0, z = 0; v < 4; v++) {
      // isolates the 4 vertices of the face
      const t_VERTEX *pVx = m_vertex + cubeFaces[noSide][v];
      pSide->m_poly[v] = ofs + pVx->v;
      z += pSide->m_zt[v] = pVx->z;
   }
   pSide->m_z = z / 4;     // average distance from the center of the face
}

/////////////////////////////////////////////////////////////////////////////
