// Sides.h: interface for the CVertex, CCubeVertexList, CSide classes.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_SIDES_H__159EB2B2_7A00_4E1B_9B37_1E3F5237A090__INCLUDED_)
#define AFX_SIDES_H__159EB2B2_7A00_4E1B_9B37_1E3F5237A090__INCLUDED_

#include "CubeUtilities.h"
#include "Matrix.h"
#include <math.h>

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


///////////////////////////////////////////////////////////////////////////////
class CVertex              // vertex in 3 D
///////////////////////////////////////////////////////////////////////////////
{
   float	m_x, m_y, m_z;    // three components of the vector

public:
   CVertex(float x = 1.0f, float y = 0.0f, float z = 0.0f)
      { m_x = x, m_y = y, m_z = z; }
   CVertex(t_Mask mask, const CMatrix *pMat, float module = 1.0f);
   virtual ~CVertex() {}

   t_Mask	Pack();        // make a pack value of the vector

   inline CVertex operator-(const CVertex &y) const
      {  return CVertex(m_x - y.m_x, m_y - y.m_y, m_z - y.m_z); }

   inline float Module() const
      {  return (float)sqrt(m_x * m_x + m_y * m_y + m_z * m_z); }

   inline float x() const { return m_x; }
   inline float y() const { return m_y; }
   inline float z() const { return m_z; }

friend class CCubeVertexList;
};

/////////////////////////////////////////////////////////////////////////////
// sommet en projection 2D (position dans le plan x, y et eloignement du plan z)
typedef struct { POINT v; int z; } t_VERTEX;


///////////////////////////////////////////////////////////////////////////////
class CCubeVertexList      // list of the 8 vertices of a cube
///////////////////////////////////////////////////////////////////////////////
{
   static float m_ds3D;    // half a side length of a 3D face
//   static float m_ds2D;    // half a side length of a 2D face

   t_VERTEX m_vertex[8];   // 8 vertices of the cube

public:

   // 2 D vertices in projection (Rubik's Cube in perspective)
   CCubeVertexList(t_Mask cube, const CMatrix *pMeo, float dim, float size, float yPos);

   // vertices in 2D (Rubik's Cube in developed flat)
   CCubeVertexList(t_Mask cube, const CMatrix &Meo, float dim, float ds);

   virtual ~CCubeVertexList() {}

   // determines whether the z side is positive
   bool IsZPos(BYTE noFace) const;

   // calcul of the face
   void Side(BYTE noFace, class CSide *pSide, const CPoint &ofs) const;

};


///////////////////////////////////////////////////////////////////////////////
class CSide                // face of a cube (projection 2 D)
///////////////////////////////////////////////////////////////////////////////
{
   int     m_z, m_zt[4];   // observer distance from the face
   CPoint  m_poly[4];      // coordinates of the face
   CRect   m_box;          // box of the face
   t_Mask  m_cube;         // Cube owner of the face
   BYTE    m_noSide;       // number of the face (color)
   BYTE    m_noDscr;       // descripteur en droule  plat (idem couleur)
   CSide  *m_pSide;        // pointer to the side classified

public:
   CSide() { m_pSide = NULL; m_noSide = 0xFF; }
   virtual ~CSide() {}

   BOOL Create(CCubeVertexList &vx, BYTE noFace, t_Mask cube, BYTE noCube, CPoint org);
   BOOL Create2D(CMatrix &meo, t_Mask cube, BYTE noDscr, CPoint org, float dim, float ds);

   BOOL PtInSide(const CPoint &pos) const;

   void Draw(CDC *pDC);
   void DrawId(CDC *pDC);
   char *GetMsg(short mode = 0, const class CCubes *pCube = NULL) const;
   static char *Mode(short mode);

   inline BYTE GetCubeNo() const { return cubeNo(m_pSide->m_cube); }


   inline void Swap(CSide *pSide)
      { CSide *pTmp = m_pSide; m_pSide = pSide->m_pSide, pSide->m_pSide = pTmp; }
   inline void Restore() { m_pSide = this; }
   inline CRect GetBox() const { return m_box; }

   int Compare(CSide *pSide) const;

   static CRect GetRect(const CSide pSide[], int nbSide);

friend class CCubes;
friend class CCubeVertexList;
};


#endif // !defined(AFX_SIDES_H__159EB2B2_7A00_4E1B_9B37_1E3F5237A090__INCLUDED_)
